import copy
import math
import os
from enum import Enum
import numpy as np
import scipy
import scipy.linalg
import scipy.ndimage.filters
import scipy.optimize
import scipy.signal
from astropy.io import fits
from scipy.interpolate import UnivariateSpline, interp1d
from typing import Tuple, Union
import h5py

from ..utils.exception import *


class Spectrum(object):
    """A class representing a single spectrum."""

    """Data type for spectra."""
    dtype = np.float64

    class Mode(Enum):
        """Spectrum wavelength mode."""
        LAMBDA = 0
        LOGLAMBDA = 1
        LOG10LAMBDA = 2

    def __init__(self, spec: 'Spectrum' = None, copy_flux: bool = True, wave: np.ndarray = None,
                 wave_start: float = None, wave_count: int = None, wave_step: float = None, wave_end: float = None,
                 wave_mode: Mode = Mode.LAMBDA, flux: np.ndarray = None, valid: np.ndarray = None):
        """Initialize a new spectrum.

        Initialization follows the following pattern:
        - Is spec given?
            - Yes, copy everything from that Spectrum. Copy flux only, if copy_flux is True.
            - No, start with empty spectrum:
                - Is wave given?
                    - Yes, just copy wavelength array.
                    - No, create new wavelength grid:
                        - If no wave_count is given, calculate it from wave_start, wave_step, and wave_end. Therefore
                          those MUST be provided!
                        - Store wave_start, wave_step, wave_count, and wave_mode
                - Do we now have a valid wavelength grid?
                    - Yes:
                        - Copy flux array, if given, otherwise create empty one
                        - Copy valid array, if given, otherwise create new one (all valid, if flux was given,
                          otherwise all invalid)

        Args:
            spec: If given, initialize this spectrum from given one.
            copy_flux: Only used in combination with spec: If False, do not copy flux array.
            wave: If given, initialize wavelength with this array.
            wave_start: If given, defines first wavelength point.
            wave_count: If given, defines number of points in spectrum.
            wave_step: If given, defines wavelength step.
            wave_end: If given, defines last wavelength point.
            wave_mode: Wavelength mode (lambda or log).
            flux: If given, defines flux array.
            valid: If given, defines valid pixel mask.
        """
        
        # start empty
        self.flux = None
        self._wave_start = None
        self._wave_step = None
        self._wavelength = None
        self._wave_mode = None
        self._valid = None

        # got a ref_spec?
        if spec is not None:
            # initialize from ref_spec
            self.flux = spec.flux.astype(Spectrum.dtype) if copy_flux else None
            self._wave_start = spec.wave_start
            self._wave_step = spec.wave_step
            self._wavelength = None if spec._wavelength is None else spec._wavelength.copy()
            self._wave_mode = spec.wave_mode
            self._valid = spec.valid

        # wavelength array given?
        if wave is not None:
            # init from wavelength
            self._wave_start = wave[0]
            self._wave_step = 0.
            self._wavelength = wave
            self._wave_mode = wave_mode
            wave_count = len(wave)

        else:
            # calculate wave_count
            if wave_count is None and wave_end is not None:
                wave_count = int((wave_end - wave_start) / wave_step)

        # init
        if wave_start is not None and wave_step is not None and wave_mode is not None:
            self._wave_start = wave_start
            self._wave_step = wave_step
            self._wavelength = None
            self._wave_mode = wave_mode

        # initialize flux and valid only if we got a valid wavelength array
        if wave_step is not None:
            # if no flux is given, initialize it empty
            if flux is None:
                # create flux array and fill with NaNs
                flux = np.empty((wave_count), dtype=Spectrum.dtype)
                flux[:] = np.nan

                # if also no valid array is given, it's all invalid now
                if valid is None:
                    # fill valid with zeros
                    valid = np.zeros(flux.shape, dtype=bool)

            else:
                # if we got some flux, but no valid array, we assume all valid
                valid = np.ones(flux.shape, dtype=bool)

        # finally set flux and valid
        if flux is not None:
            self.flux = flux
        if valid is not None:
            self._valid = valid

    @property
    def wave(self) -> np.ndarray:
        """Returns the wavelength array for this spectrum and creates it first, if necessary.

        Returns:
            Wavelength array.
        """

        if self._wavelength is None:
            # create new wavelength array
            length = len(self.flux) * self._wave_step - 0.5 * self._wave_step
            self._wavelength = np.arange(self._wave_start,
                                         self._wave_start + length,
                                         self._wave_step)
        return self._wavelength

    def pixel_borders(self) -> list:
        """Returns a list containing the borders of the pixels, always one entry more than number of wavelength points.

        Returns:
            List of pixel borders.
        """

        # get wave array
        wave = self.wave

        # get central positions between pixels
        centres = 0.5 * (wave[1:] + wave[:-1])

        # add first and last border
        centres = [2. * wave[0] - centres[0]] + list(centres)
        centres = list(centres) + [2. * wave[-1] - centres[-1]]

        # return
        return centres

    def pixel_sizes(self, scale: float = 1.) -> list:
        """Returns list of (left,right) tuples for each pixel in spectrum containing its left and right border.
        The size of the pixels can be scales using the scale parameter.

        Args:
            scale: Scale the pixel sizes with this factor.

        Returns:
            List of pixel sizes.
        """

        # get borders and calculate sizes
        borders = self.pixel_borders()
        sizes = list(zip(borders[:-1], borders[1:]))

        # scale?
        if scale == 1.:
            return sizes
        else:
            # scale left and right by (1-scale)/2
            s = (1. - scale) / 2.
            return [(l + (r - l) * s, r - (r - l) * s) for l, r in sizes]

    @property
    def valid(self) -> np.ndarray:
        """Returns the mask of valid pixels."""
        return self._valid

    @valid.setter
    def valid(self, v: np.ndarray):
        """Sets the mask of valid pixels. Must be of same length as flux."""

        # check length
        if len(v) != len(self.flux):
            raise ValueError("Length of valid array must be equal to "
                             "length of flux array.")

        # set it
        self._valid = v

    def __getitem__(self, idx: int):
        """Returns a single value from the flux array.

        Args:
            idx: Index in flux array.

        Returns:
            Flux at given index.
        """
        return self.flux[idx]

    def __setitem__(self, idx: int, value: float):
        """Sets a single value in the flux array,

        Args:
            idx: Index in flux array.
            value: Value to set.

        Returns:
            Flux at given index.
        """
        self.flux[idx] = value

    @property
    def wave_start(self) -> float:
        """Wavelength of first point in spectrum."""
        return self._wave_start

    @wave_start.setter
    def wave_start(self, v: float):
        """Sets a new start wavelength and invalidates wavelength array.

        Args:
            v: New start wavelength.
        """
        self._wave_start = v
        self._wavelength = None

    @property
    def wave_step(self) -> float:
        """Wavelength step."""
        return self._wave_step

    def __len__(self) -> int:
        """Number of pixels in spectrum."""
        return len(self.flux)

    def __iter__(self):
        """Return an iterator over the flux."""
        return iter(self.flux)

    @property
    def wave_mode(self) -> Mode:
        """Mode of spectrum."""
        return self._wave_mode

    def redshift(self, vrad: float):
        """Shift the spectrum by a given radial velocity.

        Args:
            vrad: Radial velocity in km/s to shift by.
        """

        # which mode?
        if self._wave_mode == Spectrum.Mode.LAMBDA:
            # wavelength in AA means a multiplication
            self._wavelength = self.wave * (1. + vrad / 299792.458)
            # and non-constant step size
            self._wave_step = 0
        else:
            # wavelength in log domain is simpler
            self._wavelength = self.wave + math.log(1. + vrad / 299792.458)
            self._wave_start = self._wavelength[0]

    def resample(self, spec: 'Spectrum' = None, wave: np.ndarray = None, wave_start: float = None,
                 wave_count: int = None, wave_step: float = None, wave_end: float = None, vrad: float = 0.,
                 linear: bool = False, flux_conserve: bool = False, fill_value: float = np.nan) -> 'Spectrum':
        """Resamples the spectrum to a new grid.

        Parameters defining the wavelength grid are just passed to the constructor of Spectrum.

        Args:
            spec: If given, initialize this spectrum from given one.
            wave: If given, initialize wavelength with this array.
            wave_start: If given, defines first wavelength point.
            wave_count: If given, defines number of points in spectrum.
            wave_step: If given, defines wavelength step.
            wave_end: If given, defines last wavelength point.
            vrad: In the process, do a radial velocity shift.
            linear: Do a linear interpolation instead of using a cubic spline (faster!).
            flux_conserve: Conserve flux instead of intensity.
            fill_value: Value to use for extrapolations.

        Returns:
            Resampled spectrum.
        """

        # no flux?
        if len(self.flux) < 2:
            return

        # get old wavelength array
        cur_wave = self.wave.copy()

        # radial velocity?
        if self._wave_mode == Spectrum.Mode.LAMBDA:
            cur_wave[~np.isnan(cur_wave)] *= 1. + vrad / 299792.458
        else:
            cur_wave[~np.isnan(cur_wave)] += math.log(1. + vrad / 299792.458)

        # create output spectrum
        output = self.__class__(spec=spec, wave=wave, wave_start=wave_start, wave_count=wave_count,
                                wave_step=wave_step, wave_end=wave_end, wave_mode=self.wave_mode)

        # set flux to NaN
        output.flux[:] = np.nan

        # do we have any NaNs in wave or flux?
        parts = []
        if np.any(np.isnan(cur_wave)) or np.any(np.isnan(self.flux)):
            # split into several sub spectra
            start_index = None
            for i in range(len(self.flux)):
                if np.isnan(cur_wave[i]) or np.isnan(self.flux[i]):
                    # found a NaN value
                    if start_index is not None:
                        # extract part
                        parts.append((start_index, self.extract_index(start_index, i)))
                        start_index = None
                else:
                    # start a new part
                    if start_index is None:
                        start_index = i

            # add last part?
            if start_index is not None:
                parts.append((start_index, self.extract_index(start_index, len(self.flux))))

        else:
            # single part
            parts.append((0, self))

        # loop all parts
        for start_index, part in parts:
            # interpolate
            ip = Spectrum._resample(part, output.wave, flux_conserve=flux_conserve,
                                    linear=linear, fill_value=fill_value)

            # copy all values in range of this part
            output.flux[start_index:start_index+len(ip)] = ip[start_index:start_index+len(ip)]

        # return result
        return output
    
    @staticmethod
    def _resample(spec: 'Spectrum', wave: np.ndarray, flux_conserve: bool = False, linear: bool = False,
                  fill_value: float = np.nan) -> np.ndarray:
        """Does the actual resampling of a spectrum.

        Args:
            spec: Spectrum to resample.
            wave: Output wavelength array.
            flux_conserve: If True, conserves flux instead of intensity.
            linear: If True, uses linear interpolation instead of a cubic spline.
            fill_value: Fill value for extrapolation.

        Returns:
            Resampled flux
        """

        # interpolate
        if not flux_conserve:
            # do interpolation, force linear if less than 3 points, return Nones if less than 2 points
            if len(spec) < 2 or len(wave) < 2:
                return [None] * len(wave)
            elif linear or len(spec) < 3:
                # interpolate linearly
                interpol_flux = interp1d(x=spec.wave, y=spec.flux, kind='linear',
                                         bounds_error=False, fill_value=fill_value)
                return interpol_flux(wave)

            else:
                # interpolate using a spline
                interpol_flux = UnivariateSpline(spec.wave, spec.flux, k=2, s=0, ext=1)
                new_flux = interpol_flux(wave)

                # since UnivariateSpline doesn't support a fill value, do a workaround
                new_flux[new_flux == 0] = fill_value

                #return it
                return new_flux

        else:
            # old borders
            old_borders = np.empty((len(spec.flux) + 1))
            old_borders[1:-1] = (spec.wave[1:] + spec.wave[0:-1]) * 0.5
            old_borders[0] = spec.wave[0] + spec.wave[1] - old_borders[1]
            old_borders[-1] = spec.wave[-2] - spec.wave[-1] + old_borders[-2]

            # new borders
            new_borders = np.empty((len(wave) + 1))
            new_borders[1:-1] = (wave[1:] + wave[0:-1]) * 0.5
            new_borders[0] = wave[0] + wave[1] - new_borders[1]
            new_borders[-1] = wave[-2] - wave[-1] + new_borders[-2]

            # integrate
            integr = np.concatenate([[0], np.cumsum(spec.flux)])

            # interpolate
            interpol_flux = scipy.interpolate.interp1d(x=old_borders, y=integr, kind='linear',
                                                       bounds_error=False, fill_value=1.)
            interp = interpol_flux(new_borders)

            # differentiate
            return interp[1:] - interp[0:-1]

    def resample_const(self, step: float = 0, flux_conserve: bool = False) -> 'Spectrum':
        """Resample spectrum to constant wavelength step size. If none is given, the lowest step size in
        existing grid is used.

        Args:
            step: New step size.
            flux_conserve: Whether or not to conserve flux. Otherwise intensity is conserved.

        Returns:
            Resampled spectrum
        """

        # nothing to do?
        if step == 0 and self._wave_step > 0:
            return self.copy(copy_flux=True)

        # filter wavelength array for NaNs
        w = np.sort(self.wave)
        w = w[~np.isnan(w)]

        # resample
        if step == 0:
            # get step sizes and discard zero values
            steps = np.array(w[1:] - w[:-1])
            steps = steps[~np.isnan(steps)]
            steps = steps[steps > 0]

            # get minimum and scale by two
            step = 0.5 * np.min(steps)

            # get number of steps
            n = math.ceil((w[-1] - w[0]) / step)

            # rescale step
            step = (w[-1] - w[0]) / n
        else:
            n = math.ceil((w[-1] - w[0]) / step)

        # finally resample
        return self.resample(wave_start=w[0], wave_count=n, wave_step=step, flux_conserve=flux_conserve)

    def add_noise(self, sigma: float):
        """Add gaussian noise with given sigma.

        Args:
            sigma: Sigma of noise to add.
        """
        if sigma == 0.:
            return
        for i in range(len(self.flux)):
            self.flux[i] += np.random.normal(0., self.flux[i] * sigma)

    def add_noise_snr(self, snr: float):
        """Add gaussian noise with a given S/N ratio.

        Args:
            snr: S/N ratio of noise to add.
        """
        if snr == 0.:
            return
        self.add_noise(1. / snr)

    def smooth(self, fwhm: float = None, sigma: float = None):
        """Smooth spectrum with a Gaussian kernel. Either FWHM or sigma should be given.

        Args:
            fwhm: FWHM of gaussian kernel.
            sigma: Sigma of gaussian kernel.
        """

        # smoothing not supported on irregular grids
        if self._wave_step == 0.:
            raise ValueError("Smoothing not supported for unbinned spectra.")

        # calculate sigma
        if sigma is None:
            # 2.35...=(2*math.sqrt(2*math.log(2)))
            sigma = fwhm / 2.3548200450309493

        # sigma is now in Angstroms, change it to pixels
        sigma /= self._wave_step

        # smooth
        self.flux = scipy.ndimage.filters.gaussian_filter(self.flux, sigma)

    def index_of_wave(self, wave: float) -> int:
        """Returns index of given wavelength in wavelength array.

        Args:
            wave: Wavelength to return index for.

        Returns:
            Index in wavelength array or None if outside range.
        """

        # get wave array
        w = self.wave

        # too small/large?
        if wave < w[0]:
            raise spexxyValueTooLowException()
        if wave > w[-1]:
            raise spexxyValueTooHighException()

        # find it
        for i in range(1, len(w) - 1):
            if 0.5 * (w[i] + w[i - 1]) <= wave < 0.5 * (w[i + 1] + w[i]):
                return i
        return None

    def indices_of_wave_range(self, w1: float, w2: float) -> Tuple[int, int]:
        """Same as index_of_wave, but optimized for ranges.

        Args:
            w1: Start wavelength for range.
            w2: End wavelength for range.

        Returns:
            Start and end index of range
        """

        # cover whole range?
        if w1 <= self.wave[0] and w2 >= self.wave[-1]:
            return 0, len(self.wave) - 1

        # get indices for w1 and w2
        try:
            i1 = self.index_of_wave(w1)
        except spexxyValueTooLowException:
            i1 = 0
        except spexxyValueTooHighException:
            return None
        try:
            i2 = self.index_of_wave(w2) + 1   # +1 in for including this pixel
        except spexxyValueTooLowException:
            return None
        except spexxyValueTooHighException:
            i2 = len(self.wave) - 1

        # mask at least 1px
        if i2 <= i1 < len(self.wave) - 1:
            i2 = i1 + 1

        # return indices
        return i1, i2

    def extract_index(self, i1: int, i2: int) -> 'Spectrum':
        """Extract spectrum in given index range.

        Args:
            i1: Start index to extract.
            i2: End index (not included) to extract.

        Returns:
            Extracted spectrum
        """
        if self._wave_step != 0.:
            return self.__class__(spec=self, flux=np.copy(self.flux[i1:i2]), wave_start=self.wave[i1],
                                  wave_step=self._wave_step, wave_mode=self._wave_mode)
        else:
            return self.__class__(spec=self, flux=np.copy(self.flux[i1:i2]), wave=np.copy(self.wave[i1:i2]),
                                  wave_mode=self._wave_mode)

    def extract(self, w1: float, w2: float) -> 'Spectrum':
        """Extract spectrum in given wavelength range.

        Args:
            w1: Start of wavelength range to extract.
            w2: End of wavelength range to extract.

        Returns:
            Extracted spectrum.
        """

        # get start and end indices
        s1 = self.index_of_wave(w1) or 0
        s2 = self.index_of_wave(w2) or 0
        if s1 < s2 + 1:
            s1 += 1

        # extract
        return self.extract_index(s1, s2)

    def vac_to_air(self):
        """Transforms spectrum from vacuum wavelengths to air wavelengths."""

        # copy wave
        wave = self.wave.copy()

        # if log, convert to lambda
        if self._wave_mode == Spectrum.Mode.LOGLAMBDA:
            wave = np.exp(wave)

        # convert vac to air
        sigma2 = (1e4 / wave) * (1e4 / wave)
        fact = 1. + 5.792105e-2 / (238.0185 - sigma2) + \
               1.67917e-3 / (57.362 - sigma2)

        # only set for lambda>2000A
        self._wavelength = wave.copy()
        self._wavelength[wave > 2000] = wave[wave > 2000] / fact[wave > 2000]
        self._wave_step = 0.

        # convert back to log, if necessary
        if self._wave_mode == Spectrum.Mode.LOGLAMBDA:
            self._wavelength = np.log(self._wavelength)

        # set wStart
        self._wave_start = self._wavelength[0]

    def mode(self, m: Mode):
        """Change mode of spectrum.

        Args:
            m: New mode of spectrum.
        """

        # convert
        if self._wave_mode == Spectrum.Mode.LAMBDA and m == Spectrum.Mode.LOGLAMBDA:
            # LAMBDA to LOG
            self._wavelength = np.log(self.wave)
        elif self._wave_mode == Spectrum.Mode.LAMBDA and m == Spectrum.Mode.LOG10LAMBDA:
            # LAMBDA to LOG10
            self._wavelength = np.log10(self.wave)
        elif self._wave_mode == Spectrum.Mode.LOGLAMBDA and m == Spectrum.Mode.LAMBDA:
            # LOG to LAMBDA
            self._wavelength = np.exp(self.wave)
        elif self._wave_mode == Spectrum.Mode.LOGLAMBDA and m == Spectrum.Mode.LOG10LAMBDA:
            # LOG to LOG10
            self._wavelength = np.log10(np.exp(self.wave))
        elif self._wave_mode == Spectrum.Mode.LOG10LAMBDA and m == Spectrum.Mode.LAMBDA:
            # LOG10 to LAMBDA
            self._wavelength = np.power(10., self.wave)
        elif self._wave_mode == Spectrum.Mode.LOG10LAMBDA and m == Spectrum.Mode.LOGLAMBDA:
            # LOG10 to LOG
            self._wavelength = np.log(np.power(10., self.wave))
        else:
            # do nothing
            return

        # set values
        self._wave_start = self._wavelength[0]
        self._wave_step = 0.
        self._wave_mode = m

    def norm_to_mean(self) -> float:
        """Normalize spectrum to mean flux.

        Returns:
            Factor the spectrum has been divided by.
        """
        norm = np.mean(self.flux)
        self.flux /= norm
        return norm

    def norm_to_one(self) -> float:
        """Normalize spectrum to an integrated value of one.

        Returns:
            Factor the spectrum has been divided by.
        """
        norm = scipy.integrate.trapz(self.flux, self.wave)
        self.flux /= norm
        return norm

    def norm_at_wavelength(self, wave: float = 5000, width: float = None) -> float:
        """Normalize spectrum so that flux at given wavelength is one. If width>0 and not None, use average of region.

        Args:
            wave: Wavelength to norm to.
            width: Width of region to average.

        Returns:
            Factor the spectrum has been divided by.
        """

        if width is None:
            # get flux at wavelength
            norm = self.flux[self.index_of_wave(wave)]
        else:
            # get average in range
            idx = self.indices_of_wave_range(wave - 0.5 * width, wave + 0.5 * width)
            norm = np.mean(self.flux[idx[0]:idx[1]])

        # divide and return
        self.flux /= norm
        return norm

    def __truediv__(self, other: Union['Spectrum', np.ndarray, float, int]) -> 'Spectrum':
        """Divide my flux with something else and return result.

        Args:
            other: Something to divide by.

        Returns:
            New spectrum with the result of the division.
        """

        # make a copy
        copy = self.copy(copy_flux=False)

        # check type
        if isinstance(other, Spectrum):
            copy.flux = self.flux / other.flux
        else:
            copy.flux = self.flux / other

        # return result
        return copy

    def __itruediv__(self, other: Union['Spectrum', np.ndarray, float, int]) -> 'Spectrum':
        """Divide my flux with something else.

        Args:
            other: Something to divide by.

        Returns:
            Input spectrum spec divided by other.
        """

        # check type
        if isinstance(other, Spectrum):
            self.flux /= other.flux
        else:
            self.flux /= other

        # return result
        return self

    def __mul__(self, other: Union['Spectrum', np.ndarray, float, int]) -> 'Spectrum':
        """Multiply my flux with something else and return result.

        Args:
            other: Something to multiply with.

        Returns:
            New spectrum with the result of the multiplication.
        """

        # make a copy
        copy = self.copy(copy_flux=False)

        # check type
        if isinstance(other, Spectrum):
            copy.flux = self.flux * other.flux
        else:
            copy.flux = self.flux * other

        # return result
        return copy

    def __imul__(self, other: Union['Spectrum', np.ndarray, float, int]) -> 'Spectrum':
        """Multiply my flux with something else.

        Args:
            other: Something to multiply with.

        Returns:
            Input spectrum spec multiplied with other.
        """

        # check type
        if isinstance(other, Spectrum):
            self.flux *= other.flux
        else:
            self.flux *= other

        # return result
        return self

    def __add__(self, other: Union['Spectrum', np.ndarray, float, int]) -> 'Spectrum':
        """Add something else to my flux and return result.

        Args:
            other: Something to add to my flux.

        Returns:
            New spectrum with the result of the addition.
        """

        # make a copy
        copy = self.copy(copy_flux=False)

        # check type
        if isinstance(other, Spectrum):
            copy.flux = self.flux + other.flux
        else:
            copy.flux = self.flux + other

        # return result
        return copy

    def __iadd__(self, other: Union['Spectrum', np.ndarray, float, int]) -> 'Spectrum':
        """Add something else to my flux.

        Args:
            other: Something to add to my flux.

        Returns:
            Input spectrum spec multiplied with other.
        """

        # check type
        if isinstance(other, Spectrum):
            self.flux += other.flux
        else:
            self.flux += other

        # return result
        return self

    def __sub__(self, other: Union['Spectrum', np.ndarray, float, int]) -> 'Spectrum':
        """Subtract something else from my flux and return result.

        Args:
            other: Something to subtract from my flux.

        Returns:
            New spectrum with the result of the subtraction.
        """

        # make a copy
        copy = self.copy(copy_flux=False)

        # check type
        if isinstance(other, Spectrum):
            copy.flux = self.flux - other.flux
        else:
            copy.flux = self.flux - other

        # return result
        return copy

    def __isub__(self, other: Union['Spectrum', np.ndarray, float, int]) -> 'Spectrum':
        """Subtract something else from my flux.

        Args:
            other: Something to subtract from my flux.

        Returns:
            Input spectrum spec with other subtracted.
        """

        # check type
        if isinstance(other, Spectrum):
            self.flux -= other.flux
        else:
            self.flux -= other

        # return result
        return self

    def estimate_snr(self, model: 'Spectrum') -> float:
        """Estimate S/N ratio for spectrum using a given model.

        Args:
            model: Model to compare with.

        Returns:
            Estimated S/N ratio
        """

        # get valid points
        v = ~np.isnan(model.flux) & ~np.isnan(self.flux)

        # get normalized residuals
        res = (self.flux[v] - model.flux[v]) / model.flux[v]

        # do some kappa-sigma-clipping
        for i in range(5):
            # get mean and standard deviation
            m = np.mean(res)
            s = np.std(res)

            # do clipping
            res = res[np.abs(res-m) < 4.*s]

        # return 1/s
        s = np.std(res)
        return 1 / s if s > 0 else None

    def copy(self, copy_flux: bool = True) -> 'Spectrum':
        """Create a copy of this spectrum.

        Args:
            copy_flux: If False, flux is not copied.

        Returns:
            Copy of this spectrum.
        """
        return copy.deepcopy(self) if copy_flux else copy.copy(self)

    def __copy__(self):
        """Create a shallow copy of this spectrum.

        Returns:
            Copy of this spectrum.
        """
        return self.__class__(spec=self, copy_flux=False)

    def __deepcopy__(self, memodict={}):
        """Create a deep copy of this spectrum.

        Returns:
            Copy of this spectrum.
        """
        return self.__class__(spec=self, copy_flux=True)

    @staticmethod
    def load(filename: str):
        """Loads a spectrum from file and guesses type.

        Args:
            filename: Name of file to load spectrum from.

        Returns:
            Loaded spectrum.
        """

        # expand vars on filename
        filename = os.path.expandvars(filename)

        # is it a FITS file?
        if filename.endswith(".fit") or filename.endswith(".FIT") or \
                filename.endswith(".fits") or filename.endswith(".FITS"):

            # get header
            hdr = fits.getheader(filename)

            # WAVE keyword?
            if "WAVE" in hdr.keys():
                # normal fits spectrum with WAVE extension
                return SpectrumFits(filename)

            elif hdr["NAXIS"] == 0:
                # no axis given? bin tabke!
                return SpectrumBinTableFITS(filename)

            else:
                # normal fits spectrum
                return SpectrumFits(filename)

        elif filename.endswith(".h5"):
            return SpectrumH5(filename)

        else:
            # no fits? try to load ascii...
            return SpectrumAscii(filename)


class SpectrumFitsHDU(Spectrum):
    """A Spectrum in a FITS HDU

    This class allows to load a spectrum from a FITS HDU and save it into one.
    """

    def __init__(self, hdu=None, hdu_list=None, primary=True, dtype=np.float32, filename: str = None, *args, **kwargs):
        """Initialize a new Spectrum from a FITS HDU

        Args:
            hdu: If given, HDU to load spectrum from.
            hdu_list: Probably necessary for loading wavelength array.
            primary: Whether the HDU should be a primary HDU. If HDU is given, this is ignored and derived automatically.
            dtype: Data type of array.
            filename: Name of file this HDU is in.
        """
        Spectrum.__init__(self, *args, **kwargs)

        # got a spec?
        if 'spec' in kwargs and isinstance(kwargs['spec'], SpectrumFitsHDU):
            # init from given spec
            spec = kwargs['spec']
            self._primary = primary
            self._hdu = spec._hdu.copy() if hdu is None and spec._hdu is not None else hdu
            self._dtype = spec._dtype if dtype is None else dtype

        else:
            # store
            self._hdu = hdu
            self._primary = primary
            self._dtype = dtype

            # init HDU
            if hdu:
                # overwrite primary, according to HDU
                self._primary = isinstance(hdu, fits.PrimaryHDU)

                # get data from HDU
                self.flux = hdu.data.astype(Spectrum.dtype)

                # get header
                hdr = hdu.header

                # do we have an extra HDU for wavelength array?
                if 'WAVE' in hdr:
                    # is there a HDU with this name?
                    if hdu_list is not None and hdr['WAVE'] in hdu_list:
                        # no file, must be HDU, check hdu_list
                        if hdu_list is None:
                            raise ValueError('No HDU list given for loading '
                                             'wavelength array.')

                        # get HDU and data
                        wave_hdu = hdu_list[hdr['WAVE']]

                    else:
                        # try to buld filename
                        wave_filename = os.path.join(os.path.dirname(filename), hdr['WAVE'])

                        # is it a file?
                        if os.path.exists(wave_filename):
                            # load from file
                            f = fits.open(wave_filename, memmap=False)
                            wave_hdu = f[0]
                            tmp = wave_hdu.data
                            f.close()

                        else:
                            # something else
                            raise ValueError('Could not load wavelength array')

                    # set data
                    self._wavelength = wave_hdu.data
                    self._wave_start = self._wavelength[0]
                    self._wave_step = 0

                    # type
                    if "CTYPE1" in wave_hdu.header.keys() and \
                            (wave_hdu.header["CTYPE1"] == "WAVE-LOG" or
                             wave_hdu.header["CTYPE1"] == "AWAV-LOG"):
                        self._wave_mode = Spectrum.Mode.LOGLAMBDA
                    else:
                        self._wave_mode = Spectrum.Mode.LAMBDA

                    # convert to Angstrom, if necessary
                    units = wave_hdu.header['CUNIT1'] if 'CUNIT1' in wave_hdu.header else 'Angstrom'
                    self._wavelength_to_angstrom(units)

                else:
                    # no, get wavelength array info
                    self._wave_start = hdr["START1"] if "START1" in hdr else hdr[
                        "CRVAL1"]
                    self._wave_step = hdr["STEP1"] if "STEP1" in hdr else hdr["CDELT1"]
                    if 'CRPIX1' in hdr and hdr['CRPIX1'] > 1:
                        self._wave_start -= (hdr['CRPIX1'] - 1) * self._wave_step
                    self._wavelength = None
                    if "CTYPE1" in hdr.keys() and \
                            (hdr["CTYPE1"] == "WAVE-LOG" or
                             hdr["CTYPE1"] == "AWAV-LOG"):
                        self._wave_mode = Spectrum.Mode.LOGLAMBDA
                    else:
                        self._wave_mode = Spectrum.Mode.LAMBDA

                    # convert to Angstrom, if necessary
                    units = hdr['CUNIT1'] if 'CUNIT1' in hdr else 'Angstrom'
                    self._wavelength_to_angstrom(units)

                # is it valid?
                self._valid = np.zeros(self.flux.shape, dtype=bool)

    def _wavelength_to_angstrom(self, units: str):
        """Convert wavelength units to Angstrom

        Args:
            units: Current unit of wavelength.
        """

        # everything okay?
        if units == 'Angstrom':
            return

        # get conversion factor
        factor = 1.
        if units == 'm':
            factor = 1e10
        else:
            raise ValueError('Unknown unit %s.' % units)

        # convert
        self._wave_start *= factor
        if self._wavelength:
            self._wavelength *= factor
        self._wave_step *= factor

    @property
    def header(self):
        """FITS header object"""
        hdu, _ = self.hdu()
        return hdu.header

    @property
    def primary(self):
        """Returns whether this HDU is supposed to be a primary HDU"""
        return self._primary

    def hdu(self):
        """Returns the updated HDU for this spectrum"""

        # get flux
        flux = self.flux.astype(self._dtype)

        # need to create HDU?
        if not self._hdu:
            # create HDU
            self._hdu = fits.PrimaryHDU(flux) if self._primary else fits.ImageHDU(flux)
        else:
            # just set data
            self._hdu.data = flux

        # do we need an extra HDU for the wavelength array?
        if self._wave_step != 0:
            # no
            wave_hdu = None

            #  set wavelength array values
            self._hdu.header["CRVAL1"] = (self._wave_start,
                                          "Wavelength of reference pixel")
            self._hdu.header["CDELT1"] = (self._wave_step,
                                          "Wavelength grid step size")
            self._hdu.header["CRPIX1"] = (1, "Reference pixel coordinates")
            self._hdu.header["WSTART"] = (self._wave_start,
                                          "Wavelength of reference pixel")
            self._hdu.header["WSTEP"] = (self._wave_step,
                                         "Wavelength grid step size")
            self._hdu.header["CUNIT1"] = "Angstrom"

            # type of array
            t = "AWAV" if self._wave_mode == Spectrum.Mode.LAMBDA \
                else "AWAV-LOG"
            self._hdu.header["CTYPE1"] = (t, "Type of wavelength grid")

        else:
            # yes, create it
            wave_hdu = fits.ImageHDU(self.wave)
            wave_hdu.name = 'WAVE'

            # type of array
            t = "AWAV" if self._wave_mode == Spectrum.Mode.LAMBDA \
                else "AWAV-LOG"
            wave_hdu.header["CTYPE1"] = (t, "Type of wavelength grid")

            # reference it
            self._hdu.header['WAVE'] = 'WAVE'

        # return new hdu
        return self._hdu, wave_hdu


class SpectrumFits(SpectrumFitsHDU):
    """A spectrum in a FITS file.

    The class extends SpectrumFitsHDU for a easier file handling.
    """

    def __init__(self, filename: str = None, extension: int = 0, *args, **kwargs):
        """Initialize a spectrum from a FITS file.

        Args:
            filename: Name of FITS file to load spectrum from.
            extension: Extension in FITS file to load from.
        """

        # store
        self._filename = filename

        # load it
        if filename:
            f = fits.open(filename, memmap=False)
            SpectrumFitsHDU.__init__(self, hdu=f[extension], hdu_list=f, filename=filename, *args, **kwargs)
            f.close()
        else:
            SpectrumFitsHDU.__init__(self, *args, **kwargs)

    @property
    def filename(self):
        return self._filename

    def save(self, filename: str = None, headers: dict = None):
        """Save spectrum to FITS file

        Args:
            filename: Name of file to save spectrum into.
            headers: Additional FITS headers.
        """

        # get HDU
        spec_hdu, wave_hdu = self.hdu()

        # add more headers?
        if headers:
            for key, value in headers.items():
                spec_hdu.header[key] = value

        # filename
        filename = filename or self._filename
        if not filename:
            raise ValueError('No filename given for saving the FITS file.')

        # create HDU list and save it
        hdu_list = fits.HDUList([spec_hdu])
        if wave_hdu:
            hdu_list.append(wave_hdu)
        hdu_list.writeto(filename, overwrite=True)

        # store filename
        self._filename = filename


class SpectrumHiResFITS(Spectrum):
    """Handles HiRes spectra from the PHOENIX library."""

    def __init__(self, filename: str = None, *args, **kwargs):
        """Loads a HiRes spectrum from file.

        Args:
            filename: Name of file to load spectrum from.
        """
        Spectrum.__init__(self, *args, **kwargs)

        # file given?
        if filename is not None:
            # get flux
            self.flux = fits.getdata(filename, 0)
            if len(self.flux) == 1 and len(self.flux[0]) > 1:
                self.flux = self.flux[0]

            # get header
            hdr = fits.getheader(filename, 0)

            # get path of filename
            path = os.path.dirname(filename)

            # get filename of wavelength file
            wave_file = path + ('/' if len(path) > 0 else "") + hdr["WAVE"]

            # get wavelength array
            self._wave_step = 0.
            self._wavelength = fits.getdata(wave_file)
            if "CTYPE1" in hdr.keys() and \
                    (hdr["CTYPE1"] == "WAVE-LOG" or hdr["CTYPE1"] == "AWAV-LOG"):
                self._wave_mode = Spectrum.Mode.LOGLAMBDA
            else:
                self._wave_mode = Spectrum.Mode.LAMBDA

            # is it valid?
            self._valid = np.zeros(self.flux.shape, dtype=bool)


class SpectrumBinTableFITS(Spectrum):
    """Handles spectra stored in FITS binary tables."""

    def __init__(self, filename: str, *args, **kwargs):
        """Loads a spectrum from a FITS binary table.

        Args:
            filename: Name of file to load spectrum from.
        """
        Spectrum.__init__(self, *args, **kwargs)

        # store
        self._filename = filename

        # load data
        if filename:
            try:
                data = fits.getdata(filename, "SPECTRUM")
                self._wavelength = data.WAVE
                self.flux = data.FLUX
            except KeyError:
                data = fits.getdata(filename, "SCI")
                self._wavelength = data.WAVELENGTH
                self.flux = data.FLUX

            # set wavelength array
            self._wave_start = self._wavelength[0]
            self._wave_step = 0

            # get array and decide on wave mode
            hdr = fits.getheader(filename)
            if "CTYPE1" in hdr.keys() and \
                    (hdr["CTYPE1"] == "WAVE-LOG" or
                     hdr["CTYPE1"] == "AWAV-LOG"):
                self._wave_mode = Spectrum.Mode.LOGLAMBDA
            else:
                self._wave_mode = Spectrum.Mode.LAMBDA

            # valid?
            self._valid = np.zeros(self.flux.shape, dtype=bool)

    def save(self, filename: str):
        """Save spectrum to FITS file

        Args:
            filename: If given, save to this filename.
        """

        # filename
        filename = filename or self._filename
        if not filename:
            raise ValueError('No filename given for saving the FITS file.')

        # create binary table
        c1 = fits.Column(name='WAVE', format='1D',
                         unit='Angstrom', array=self._wavelength)
        c2 = fits.Column(name='FLUX', format='1E',
                         unit='erg/s/cm2/Angstrom', array=self.flux)
        hdu = fits.new_table([c1, c2])
        hdu.header["EXTNAME"] = "SPECTRUM"

        # create HDU list and save it
        hdu_list = fits.HDUList([hdu])
        hdu_list.writeto(filename)

        # store filename
        self._filename = filename


class SpectrumAscii(Spectrum):
    """Handles spectra stored in ASCII files."""

    def __init__(self, filename: str = None, separator: str = ',', skip_lines: int = 0, comment: str = '#',
                 wave_column: int = 0, flux_column: int = 1, header: bool = True, *args, **kwargs):
        """Reads spectrum from ASCII file.

        Args:
            filename: Filename to load spectrum from.
            separator: Separator in file.
            skip_lines: Number of lines to skip while reading file.
            comment: Character indicating a comment line.
            header: Whether to write a header.
        """
        Spectrum.__init__(self, *args, **kwargs)

        # store
        self._filename = filename
        self._separator = separator
        self._header = header

        # load
        if filename:
            # init arrays
            wave = []
            flux = []

            # open file
            with open(filename, "r") as txt:
                # skip lines
                for i in range(skip_lines):
                    txt.readline()

                # read lines
                for line in txt:
                    if line[0] == comment:
                        continue
                    s = line.split(separator)
                    wave.append(float(s[wave_column]))
                    flux.append(float(s[flux_column]))

            # set everything
            self._wavelength = np.array(wave)
            self.flux = np.array(flux)
            self._wave_start = self._wavelength[0]
            self._wave_step = 0
            self._wave_mode = Spectrum.Mode.LOGLAMBDA \
                if self._wave_start < 10 else Spectrum.Mode.LAMBDA
            self._valid = np.zeros(self.flux.shape, dtype=bool)

    def save(self, filename: str):
        """Save spectrum to ASCII file

        Args:
            filename: If given, save to this filename.
        """

        # filename
        filename = filename or self._filename
        if not filename:
            raise ValueError('No filename given for saving the ASCII file.')

        # save it
        with open(filename, 'w') as f:
            # write header?
            if self._header:
                f.write('wavelength,flux\n')

            # write flux
            wave = self.wave
            for i in range(len(self.flux)):
                f.write("{0:f}{1:s}{2:f}\n".format(wave[i], self._separator, self.flux[i]))

        # store filename
        self._filename = filename


class SpectrumH5(Spectrum):
    """Handles spectra stored in H5 files."""

    def __init__(self, filename: str = None, *args, **kwargs):
        """Reads spectrum from H5 file.

        Args:
            filename: Filename to load spectrum from.
        """
        Spectrum.__init__(self, *args, **kwargs)

        # store
        self._filename = filename

        # load
        if filename:
            # read from H5
            fh5 = h5py.File(filename, 'r')
            wave = fh5['PHOENIX_SPECTRUM/wl'][()]
            flux = 10. ** fh5['PHOENIX_SPECTRUM/flux'][()]
            fh5.close()

            # set everything
            self._wavelength = np.array(wave)
            self.flux = np.array(flux)
            self._wave_start = self._wavelength[0]
            self._wave_step = 0
            self._wave_mode = Spectrum.Mode.LOGLAMBDA \
                if self._wave_start < 10 else Spectrum.Mode.LAMBDA
            self._valid = np.zeros(self.flux.shape, dtype=bool)

    def save(self, filename: str):
        """Save spectrum to H5 file

        Args:
            filename: If given, save to this filename.
        """
        pass

    @property
    def filename(self):
        return self._filename


__all__ = ['Spectrum', 'SpectrumAscii', 'SpectrumBinTableFITS', 'SpectrumFits', 'SpectrumFitsHDU', 'SpectrumHiResFITS',
           'SpectrumH5']


from typing import Any
from lmfit import Parameters
import numpy as np

from spexxy.object import spexxyObject
from spexxy.data import FitsSpectrum


class Component(spexxyObject):
    """Base class for all Components in spexxy."""

    """Data type for parameters"""
    dtype = np.float64

    def __init__(self, name: str, init: list = None, prefix: str = None, normalize: bool = False, *args, **kwargs):
        """Initialize a new component.

        Args:
            name: Name of component
            init: List of Init objects for initializing the component
            prefix: Prefix for parameter name when combined in other model. Automatically derived from name if None.
            normalize: Whether or not to normalize parameters to 0..1 range
        """
        spexxyObject.__init__(self, *args, **kwargs)

        # name and prefix
        self.name = name
        self.prefix = prefix if prefix is not None else name

        # weight for params fit
        self.weight = 1.

        # norm?
        self.normalize = normalize

        # parameters
        self.parameters = {}

        # create init functions
        from spexxy.init import Init
        self._init = [] if init is None else self.get_objects(init, Init, 'init')

    def __call__(self, **params) -> Any:
        """Model function, must be implemented.

        Args:
            params: Parameters to retrieve model for

        Returns:
            The model from the component
        """
        raise NotImplementedError

    @property
    def param_names(self):
        return self.parameters.keys()

    def set(self, name, **kwargs):
        """Adds a new parameter with the given name or changes values of an existing one.

        Args:
            name(str): Parameter name.
            value (float): (Initial) Value of parameter
            stderr (float): Standard deviation after fit
            vary (bool): Whether or not to vary the parameter in the fit
            min (float): Lower bound for fit (default is `-numpy.inf`, no lower bound)
            max (float): Upper bound for value (default is `numpy.inf`, no upper bound)
            kwargs: Passed to set_param_hint()
        """

        # remove prefix from name if exists
        n = len(self.prefix)
        if n > 0 and name.startswith(self.prefix):
            name = name[n:]

        # add parameter if not exists
        if name not in self.parameters:
            self.parameters[name] = {}

        # set values
        for key, val in kwargs.items():
            if key in ('value', 'stderr', 'vary', 'min', 'max'):
                self.parameters[name][key] = Component.dtype(val)

    def __getitem__(self, name: str) -> float:
        """Returns the value for an existing parameter.

        Args:
            name: Name of parameter for which to return the value

        Returns:
            Value of given parameter

        Raises:
            KeyError: If given parameter does not exist
        """
        return self.parameters[name]['value'] if 'value' in self.parameters[name] else None

    def __setitem__(self, name: str, value: float):
        """Returns the value for an existing parameter.

        Args:
            name: Name of parameter to set
            value: New value for parameter
        """
        self.set(name, value=Component.dtype(value))

    def make_params(self, **kwargs) -> Parameters:
        """Creates a Parameters object with a Parameter for each parameter of this component.

        Args:
            kwargs: Values to overwrite in parameters

        Returns:
            List of Parameters for this component
        """

        # init params
        params = Parameters()

        # remove prefixes from kwargs if exists
        n = len(self.prefix)
        overwrite = {key[n:]: value for key, value in kwargs.items() if n > 0 and key.startswith(self.prefix)}

        # loop all parameters
        for name, p in self.parameters.items():
            # copy p
            param = dict(p)

            # overwrite value?
            if name in overwrite:
                param['value'] = overwrite[name]

            # default values
            for key, val in {'value': None, 'vary': True, 'min': -np.inf, 'max': np.inf}.items():
                if key not in param:
                    param[key] = val

            # remove stderr
            for k in ['stderr']:
                if k in param:
                    del param['stderr']

            # normalize?
            if self.normalize:
                param['value'] = self.norm_param(name, param['value'])
                param['min'] = 0
                param['max'] = 1

            # got vary parameter?
            if 'vary' in param:
                # parse it to bool
                param['vary'] = bool(param['vary'])

                # if it's False, set min/max to +-inf to avoid "min == max" errors
                if not param['vary']:
                    param['min'] = -np.inf
                    param['max'] = np.inf

            # add it to params
            params.add(self.prefix + name, **param)

        # return it
        return params

    def parse_params(self, params: Parameters):
        """Loop all Parameters in a Parameters object and set the values of this component accordingly.

        Args:
            params: Parameters objects, usually return from a lmfit optimization
        """

        # length of prefix
        n = len(self.prefix)

        # loop parameters
        for name, param in params.items():
            # does prefix match?
            if n > 0 and name.startswith(self.prefix):
                # remove prefix
                name = name[n:]

                # does it exist?
                if name in self.parameters:
                    # de-normalize?
                    if self.normalize:
                        value, stderr = self.denorm_param(name, param.value, param.stderr)
                    else:
                        value, stderr = param.value, param.stderr

                    # set it
                    self.set(name, value=value, stderr=stderr)
                    
    def norm_param(self, name: str, value: float) -> float:
        """Normalize the value of the parameter with the given name to 0..1 range defined by its min/max.

        Args:
            name: Name of parameter to normalize.
            value: Value to normalize.

        Returns:
            Normalized value.
        """
        param = self.parameters[name]
        return (value - param['min']) / (param['max'] - param['min'])

    def denorm_param(self, name: str, value: float, stderr: float = None) -> (float, float):
        """De-Normalize the value of the parameter with the given name to its real value.

        Args:
            name: Name of parameter to de-normalize.
            value: Value to normalize.
            stderr: If given, standard deviation of value.

        Returns:
            Normalized value and its standard deviation, if given.
        """
        param = self.parameters[name]
        val = value * (param['max'] - param['min']) + param['min']
        std = None if stderr is None else stderr * (param['max'] - param['min'])
        return val, std

    def init(self, filename: str):
        """Calls all Init objects with the given filename in order to initialize this component.

        Args:
            filename: File to initialize with, may be optional for some Init objects
        """
        for fcn in self._init:
            fcn(self, filename)

    def write_results_to_file(self, fits_file: FitsSpectrum):
        """Write results of this component into a given SpectrumFile

        Args:
            fits_file: Opened FitsSpectrum to write results into
        """

        # get results object
        params = fits_file.results(self.prefix)

        # loop parameter names
        for name, param in self.parameters.items():
            # write results into results object
            vary = 'vary' not in param or param['vary'] is True
            params[name] = [param['value'], param['stderr'] if vary else None]


__all__ = ['Component']



class spexxyException(Exception):
    def __init__(self, message=""):
        self.message = message


class spexxyValueTooLowException(spexxyException):
    pass


class spexxyValueTooHighException(spexxyException):
    pass



from astropy.io import fits
from typing import Union, Tuple, Any


class ResultsFITS(object):
    """Handles results in the primary header of a FITS file."""

    def __init__(self, hdu: fits.PrimaryHDU, namespace: str, prefix: str = 'ANALYSIS'):
        """Initialize a new results set or loads it from a HDU

        Args:
            hdu: HDU of FITS file to store/retrieve results into/from.
            namespace: Namespace for results.
        """
        self._hdu = hdu
        self._prefix = [prefix.upper()]
        self._namespace = self._prefix + namespace.upper().split()
        self._data = {}

        # load from file
        self._load()

    def __getitem__(self, key: str) -> Tuple[float, float]:
        """Returns a single result as (value,error) pair.

        Args:
            key: Name of result.

        Returns:
            Result as (value,error) pair
        """

        # if key doesn't exist, return None
        if key.upper() not in self._data:
            return None

        # get value from dict
        return self._data[key.upper()]

    def __setitem__(self, key: str, value: Union[Tuple, float, str]):
        """Set a new result.

        Args:
            key: Name of result.
            value: Either a value or a list-like value/error pair.
        """

        # just a value or value/error list?
        if hasattr(value, '__iter__') and not isinstance(value, str):
            # if it is a list, we need two entries for value and error
            if len(value) == 2:
                self._data[key.upper()] = value
            else:
                raise ValueError("Value and error must be given.")
        else:
            # just set value with no error
            self._data[key.upper()] = [value, None]

        # save results back to HDU
        self._save()

    def __delitem__(self, key: str):
        """Delete result from set.

        Args:
            key: Name of result.
        """

        # does key exist?
        if key in self._data:
            # delete entry
            del self._data[key]

            # save changes back to HDU
            self._save()

    def __len__(self):
        """Number of results in set."""
        return len(self._data)

    def __iter__(self):
        """Return iterator over all result names."""
        return iter(self._data.keys())

    def __contains__(self, item):
        """Whether results set contains the given item."""
        return item.upper() in self._data

    def keys(self):
        """Returns list of keys in result set."""
        return self._data.keys()

    @property
    def data(self):
        """Returns results set as dict."""
        return self._data

    def _load(self):
        """Load results set from HDU."""

        # init
        self._data = {}
        s = len(self._namespace)

        # get header
        hdr = self._hdu.header.cards

        # loop all entries
        for h in hdr:
            # split keyword
            key = str(h.keyword).split()

            # is this the namespace we want?
            if key[:s] == self._namespace:
                # get name
                name = " ".join(key[s:-1] if key[-1] == "ERR" else key[s:])

                # check data entry
                if name not in self._data:
                    self._data[name] = [None, None]

                # set it
                if key[-1] == "ERR":
                    self._data[name][1] = h.value
                else:
                    self._data[name][0] = h.value

    def _update_header(self, key: str, value: Any):
        """Update HDU header keyword.

        Args:
            key: Name of header keyword.
            value: New value for header entry.
        """

        # does keyword exist?
        if key.upper() in list(self._hdu.header.keys()):
            self._hdu.header[key.upper()] = value
        else:
            self._hdu.header["HIERARCH " + key.upper()] = value

    def _save(self):
        """Write result set back to HDU header."""

        # number of text messages written
        num_msg = 1

        # first remove all keys of this namespace
        hdr = self._hdu.header
        for key in list(hdr.keys()):
            # base string for this key
            base = "%s" % " ".join(self._namespace)
            if str(key.strip()).startswith(base):
                del hdr["HIERARCH " + key.strip()]

        # loop all data
        for key in self._data.keys():
            # base string for this key
            base = str("%s %s" % (" ".join(self._namespace), key))

            # write value
            if type(self._data[key][0]) is str:
                # create new MSG entry, if value is a string
                self._update_header(base, "MSG%04d" % num_msg)
                hdr["MSG%04d" % num_msg] = self._data[key][0]
                num_msg += 1
            else:
                # just write the value
                self._update_header(base, self._data[key][0])

            # write error
            if self._data[key][1] is not None:
                self._update_header(base + " ERR", self._data[key][1])


__all__ = ['ResultsFITS']



import logging
import argparse
import os
import numpy as np
from astropy.io import fits

from spexxy.grid import Grid
from spexxy.data import LSF, Spectrum, SpectrumFits

log = logging.getLogger(__name__)


def add_parser(subparsers):
    # create parser
    parser = subparsers.add_parser('convert', help='Convert a grid from one type to another.',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('input', type=str, help='Input grid')
    parser.add_argument('--norm-to-mean', action='store_true', help='Norm input spectra to mean.')
    subparsers = parser.add_subparsers()

    # to files grid
    parser_files = subparsers.add_parser('files', help='Convert to files grid')
    parser_files.set_defaults(func=_convert_to_files)
    parser_files.add_argument('outdir', type=str, help='Output path')
    parser_files.add_argument('-c', '--csv', type=str, help='CSV filename', default='grid.csv')
    pattern = 'Z-{FeH:.2f}.Alpha={Alpha:.2f}/lte{Teff:05.0f}{logg:+.2f}{FeH:+.2f}.Alpha={Alpha:.2f}.fits'
    parser_files.add_argument('-f', '--filenames', type=str, help='Filename pattern', default=pattern)

    # to fits grid
    parser_fits = subparsers.add_parser('fits', help='Convert to fits grid')
    parser_fits.set_defaults(func=_convert_to_fits)
    parser_fits.add_argument('output', type=str, help='Output file')


def _convert_to_files(args):
    # load grid
    grid = Grid.load(args.input)

    # does output directory exist?
    if not os.path.exists(args.outdir):
        os.makedirs(args.outdir)

    # open CSV file
    with open(os.path.join(args.outdir, args.csv), 'w') as csv:
        # write header
        csv.write('Filename,' + ','.join(grid.axis_names()) + '\n')

        # loop all parameters
        all_params = grid.all()
        for i, params in enumerate(all_params, 1):
            log.info('(%d/%d) %s...', i, len(all_params), params)

            # get spectrum
            spec = SpectrumFits(spec=grid(params))

            # norm?
            if args.norm_to_mean:
                spec.norm_to_mean()

            # get filename
            p = dict(zip(grid.axis_names(), params))
            filename = args.filenames.format(**p)

            # directory
            spec_filename = os.path.join(args.outdir, filename)
            spec_path = os.path.dirname(spec_filename)
            if not os.path.exists(spec_path):
                os.makedirs(spec_path)

            # save spectrum
            spec.save(spec_filename)

            # add to csv
            csv.write(filename + ',' + ','.join([str(p) for p in params]) + '\n')


def _convert_to_fits(args):
    # load grid and get all parameters
    grid = Grid.load(args.input)
    all_params = grid.all()

    # load first spectrum and get its length
    spec = grid(all_params[0])
    nwave = len(spec)

    # Now we want to write a probably very large FITS file. There are instructions for that:
    # https://docs.astropy.org/en/stable/generated/examples/io/skip_create-large-fits.html
    # Unfortunately, it only describes, how to create an empty file, not how to fill it without loading it again,
    # which then loads the whole file into memory. So, in the end it still consumes all memory, but at least we create
    # the file beforehand.
    # TODO: Try to fix this.

    # create a dummy header
    data = np.zeros((100, 100), dtype=np.float32)
    hdu = fits.PrimaryHDU(data=data)
    header = hdu.header
    while len(header) < 10:
        header.append()  # Adds a blank card to the end

    # reset size, set some headers and write to file
    header['NAXIS1'] = nwave
    header['NAXIS2'] = len(all_params)
    header['CRVAL1'] = header['WSTART'] = (spec.wave_start, 'Wavelength at reference pixel')
    header['CDELT1'] = header['WSTEP'] = (spec.wave_step, 'Wavelength grid step size')
    header['CRPIX1'] = (spec.wave_step, 'Reference pixel coordinates')
    header['CTYPE1'] = ('AWAV' if spec.wave_mode == Spectrum.Mode.LAMBDA else 'AWAV-LOG', 'Type of wavelength grid')
    header.tofile(args.output, overwrite=True)

    # grow file to required size
    shape = tuple(header['NAXIS{0}'.format(ii)] for ii in range(1, header['NAXIS'] + 1))
    with open(args.output, 'rb+') as fobj:
        fobj.seek(len(header.tostring()) + (np.product(shape) * np.abs(header['BITPIX'] // 8)) - 1)
        fobj.write(b'\0')

    # now we can open the file for writing
    output = fits.open(args.output, mode='update', memmap=False)

    # loop all parameters
    all_params = grid.all()
    for i, params in enumerate(all_params, 1):
        log.info('(%d/%d) %s...', i, len(all_params), params)

        # get spectrum
        spec = SpectrumFits(spec=grid(params))

        # norm?
        if args.norm_to_mean:
            spec.norm_to_mean()

        # write to FITS file
        output[0].data[i - 1, :] = spec.flux

    # transpose parameter list, so that we have columns instead of rows
    params = list(map(list, zip(*all_params)))

    # create columns
    params_cols = [fits.Column(name=name, format='E', array=data) for name, data in zip(grid.axis_names(), params)]

    # add table with parameters
    hdu_table = fits.BinTableHDU.from_columns(params_cols)
    hdu_table.name = 'PARAMS'
    output.append(hdu_table)

    # flush file and close
    output.flush()
    output.close()



import os

import scipy
from astropy.io import fits

from spexxy.data import Filter, Spectrum
from spexxy.grid import FilesGrid


# define list of FITS keywords to copy to new files
KEYWORDS = ['PHXTEFF', 'PHXLOGG', 'PHXM_H', 'PHXDUST', 'PHXXI_L', 'PHXXI_M', 'PHXXI_N', 'PHXMASS', 'PHXREFF', 'PHXLUM',
            'PHXMXLEN', 'PHXBUILD', 'PHXVER', 'DATE', 'PHXEOS']


def add_parser(subparsers):
    # init parser
    parser = subparsers.add_parser('limbdark', help='Apply filter to specific intensities file(s)')
    parser.set_defaults(func=run)

    # parameters
    parser.add_argument('input', help='Spectrum or grid file (in --grid mode)', type=str)
    parser.add_argument('output', help='Output filename or directory (in --grid mode)', type=str)
    parser.add_argument('filter', help='Filter name or filename', type=str)
    parser.add_argument('-g', '--grid', help='Process all files in grid file given as input', action='store_true')


def run(args):
    cld = ConvolveLimbDark(args.input, args.output, args.filter, args.grid)
    cld()


class ConvolveLimbDark:
    def __init__(self, input: str, output: str, fltr: str, grid_mode: bool = False):
        """Initialises object

        Args:
            input: Input filename/grid
            output: Output filename/directory
            fltr: Filter name
            grid_mode: Whether input is a grid
        """

        # store
        self.input = input
        self.output = output
        self.grid_mode = grid_mode

        # find filter
        self.filter = Filter(fltr)
        if self.filter.filename is None:
            raise ValueError('Could not find filter: ' + fltr)
        self.filter_resampled = False

    def __call__(self):
        """Process input file/grid"""

        # grid mode or single spectrum?
        if self.grid_mode:
            # load grid
            grid = FilesGrid(self.input)

            # process it
            self.process_grid(grid)

        else:
            # process single file
            self.process_single(self.input, self.output)

    def process_grid(self, grid):
        """Convolves all specific intensity spectra in the given grid.

        Args:
            grid: Grid to convolve.
        """

        # loop all entries in grid
        for params in grid.all():
            # get input filename
            in_filename = grid.filename(params)

            # create output filename
            out_filename = os.path.join(self.output, grid.filename(params, absolute=False))

            # process file
            self.process_single(in_filename, out_filename)

    def process_single(self, input, output):
        """Convolves a specific intensities spectrum with the given filter and stores result."""

        # load data and mus
        mu = fits.getdata(input, 'MU')
        data, hdr = fits.getdata(input, header=True)
        wave_start = hdr['CRVAL1']
        wave_step = hdr['CDELT1']

        # loop all mus
        intensities = []
        for k in range(len(mu)):
            # create spectrum
            spec = Spectrum(flux=data[k, :], wave_start=wave_start, wave_step=wave_step)

            # resample filter?
            if not self.filter_resampled:
                self.filter.resample(spec=spec, fill_value=0.)
                self.filter_resampled = True

            # apply filter
            intensity = self.filter.integrate(spec)
            intensities.append(intensity)

        # normalize to 1 at centre
        # intensities /= intensities[-1]

        # integrate I(mu)*mu
        H = scipy.integrate.trapz(intensities * mu, mu)

        # scale limb darkening as I(mu)/2H
        intensities /= 2 * H

        # create HDU
        hdu_primary = fits.PrimaryHDU(intensities)
        hdu_primary.header['WAVE'] = 'MU'

        # set headers
        for key in KEYWORDS:
            hdu_primary.header[key] = hdr[key]

        # HDU for MUs
        hdu_mu = fits.ImageHDU(mu)
        hdu_mu.name = 'MU'

        # path exists?
        path = os.path.dirname(output)
        if not os.path.exists(path):
            os.makedirs(path)

        # store it
        fits.HDUList([hdu_primary, hdu_mu]).writeto(output, overwrite=True)



import logging
import os
import sys
import typing


class spexxyObject(object):
    """Base class for all objects in spexxy."""

    def __init__(self, objects: dict = None, log: logging.Logger = None, *args, **kwargs):
        """Initializes a new object.

        Args:
            objects: Dictionary containing other objects.
            log: Logging instance to use.
        """
        self.objects = {} if objects is None else objects
        self._log = log

    @property
    def log(self) -> logging.Logger:
        """Get logger for this object.

        Returns:
            Logger to use for this object.
        """
        return self._log if self._log is not None else logging.getLogger()

    def _get_object_from_group(self, name: str, klass, group: str) -> 'spexxyObject':
        """Returns an object from a group.

        Args:
            name: Name of object to retrieve from group.
            klass: Class of object to retrieve from group.
            group: Group to retrieve object from.

        Returns:
            Object with given name and class from the given group.

        Raises:
            ValueError: If object cannot be found.
        """

        # check existence
        if group in self.objects and name in self.objects[group]:
            # get obj and check type
            obj = self.objects[group][name]
            if isinstance(obj, klass):
                return obj
        else:
            # nothing found
            raise ValueError('Object "%s" of class "%s" not found in group "%s".' % (name, klass.__name__, group))

    def _find_name_in_group(self, group: str) -> str:
        """Finds a new unused name in the given group.

        Args:
            group: Group to create a name for.

        Returns:
            New unique name for group-
        """

        # create group if required
        if group not in self.objects:
            self.objects[group] = {}

        # check group name itself
        if group not in self.objects[group]:
            return group

        # otherwise count up
        i = 1
        while True:
            name = '%s-%d' % (group, i)
            if name not in self.objects[group]:
                return name
            i += 1

    def _create_object(self, name: str, definition: dict, klass, group: str, log: logging.Logger = None) \
            -> 'spexxyObject':
        """Creates a single object from the given definition.

        Args:
            name: Name for new object to create.
            definition: Dictionary containing the description of the object to create, especially a "class" element.
            klass: Class the new object should have.

        Returns:
            New object created from definition.

        Raises:
            ValueError: If newly created class is not of given class.
        """

        # create object
        obj = create_object(definition, log=log, objects=self.objects)

        # check type
        if not self._is_of_classes(obj, klass):
            raise ValueError('Newly created object "%s" is of type "%s", should be one of %s.' %
                             (name, obj.__class__.__name__, [k.__name__ for k in klass]))

        # add to objects dict
        if group not in self.objects:
            self.objects[group] = {}
        self.objects[group][name] = obj

        # return it
        return obj

    def _is_of_classes(self, obj: object, klass: typing.Union[typing.Type, typing.List[typing.Type]]):
        """Checks, whether

        Args:
            obj: Object to check
            klass: Single type or list of types

        Returns:

        """

        # make it a list
        if not isinstance(klass, list):
            klass = [klass]

        # loop all
        for k in klass:
            if isinstance(obj, k):
                # found it
                return True
        else:
            # nothing found
            return False

    def get_objects(self, definition: typing.Union[object, str, list, dict], klass, group: str,
                    log: logging.Logger = None, single: bool = False) -> typing.Union[list, 'spexxyObject', None]:
        """Get either a single object or a list of objects from the given definition

        Args:
            definition: Definition of a single object or a list of objects or an object itself.
            klass: Expected class for object.
            group: Group to get object from or put it into.
            log: Logger for newly created object.
            single: Always return single element or None instead of list.

        Returns:
            Either a single object or a list of objects or None

        Raises:
            ValueError: If object could not be retrieved/created.
        """

        # if no logger is given, use mine
        if log is None:
            log = self._log

        # init list of objs
        objs = []

        # decide on what we got in definition
        if definition is None:
            # just nothing
            pass

        elif self._is_of_classes(definition, klass):
            # got the correct type from the beginning!
            objs = [definition]

        elif isinstance(definition, str):
            # definition must be a name of an object in the given group
            objs = [self._get_object_from_group(definition, klass, group)]

        elif isinstance(definition, list):
            # definition can be a list of strings with names of objects in the given group or a list of defs
            # loop all definitions and get them
            for defn in definition:
                if isinstance(defn, str):
                    obj =self._get_object_from_group(defn, klass, group)
                elif isinstance(defn, dict) and 'class' in defn:
                    # single def, find a good name and create it
                    name = self._find_name_in_group(group)
                    obj = self._create_object(name, defn, klass, group, log=log)
                else:
                    raise ValueError('Invalid definition.')

                # add it
                objs.append(obj)

        elif isinstance(definition, dict):
            # is this just a single def or multiple?
            if 'class' in definition:
                # single def, find a good name and create it
                name = self._find_name_in_group(group)
                objs = [self._create_object(name, definition, klass, group, log=log)]

            else:
                # definition must be a dict of name->definition pairs for objects to create
                if any(['class' not in d for d in definition.values()]):
                    raise ValueError('Dict must contain name->definition pairs.')

                # create them
                objs = [self._create_object(name, defn, klass, group, log=log)
                        for name, defn in definition.items()]

        else:
            raise ValueError('Unknown type for object definition.')

        # single or list?
        return (None if len(objs) == 0 else objs[0]) if single else objs


def get_class_from_string(class_name: str) -> typing.Type:
    """Take a class name as a string and return the actual class

    Args:
        class_name: Name of class.

    Returns:
        Actual class.
    """

    # split parts of class name, i.e. modules and class
    parts = class_name.split('.')

    # join module name
    module_name = ".".join(parts[:-1])

    # import module
    sys.path.insert(0, os.getcwd())
    cls = __import__(module_name)

    # fetch class and return it
    for comp in parts[1:]:
        cls = getattr(cls, comp)
    return cls


def create_object(config: dict, log: logging.Logger = None, *args, **kwargs) -> spexxyObject:
    """Create a new object from a dict.

    Args:
        config: Dictionary with a "class" element to create object from.
        log: Logger to use for new object.

    Returns:
        New object created from config.

    Raises:
        ValueError: Cannot copy config dictionary or no class name given.
    """

    # copy config
    try:
        cfg = dict(config)
    except ValueError:
        log.error('Cannot copy dict: %s', config)
        raise

    # get class name
    class_name = cfg.pop('class', None)
    if class_name is None:
        raise ValueError('No class name given.')

    # create class
    cls = get_class_from_string(class_name)

    # create object
    return cls(*args, **kwargs, **cfg, log=log)


__all__ = ['spexxyObject', 'create_object']



import copy
import numpy as np
from typing import Tuple, Any, List

from .grid import Grid, GridAxis
from ..data import Spectrum


class FerreGrid(Grid):
    def __init__(self, filename: str):
        """Load a Ferre grid.
        See: https://github.com/callendeprieto/ferre/blob/master/docs/ferre.pdf

        Args:
            filename: Name of file to load.
        """

        # init
        self._filename = filename
        self._n_of_dim = 0
        self._n_p = []
        self._labels = []
        self._llimits = []
        self._steps = []
        self._npix = None
        self._wave = []
        self._logw = None
        self._spec_start = None
        self._line_length = None
        self._line_pos = []
        self._params_line = {}

        # load grid
        self._load_grid()
        self._map_lines()

        # build axes
        axes = []
        for i in range(self._n_of_dim):
            values = [self._llimits[i] + k * self._steps[i] for k in range(self._n_p[i])]
            axes.append(GridAxis(name=self._labels[i],
                                 values=values,
                                 min=values[0],
                                 max=values[-1]))

        # init grid
        Grid.__init__(self, axes)

    def _load_grid(self):
        # open file
        with open(self._filename) as f:
            # check filetype
            if f.readline().strip() != '&SYNTH':
                raise ValueError('Given file is not a Ferre grid.')

            # read header
            for line in f:
                # end of header?
                if line.strip() == '/':
                    break

                # split by =
                p = line.index('=')
                key = line[:p].strip()
                value = line[p + 1:].strip()

                # remove quotes
                if len(value) > 0 and value[0] in ['"', "'"]:
                    value = value[1:-1]

                # set values
                if key == 'N_OF_DIM':
                    self._n_of_dim = int(value)
                    self._labels = [''] * self._n_of_dim
                    self._llimits = [0] * self._n_of_dim
                    self._steps = [0] * self._n_of_dim
                elif key == 'N_P':
                    self._n_p = [int(n) for n in value.split()]
                elif key.startswith('LABEL'):
                    i = int(line[line.index('(') + 1:line.index(')')])
                    self._labels[i - 1] = value
                elif key == 'LLIMITS':
                    self._llimits = [float(f) for f in value.split()]
                elif key == 'STEPS':
                    self._steps = [float(f) for f in value.split()]
                elif key == 'NPIX':
                    self._npix = int(value)
                elif key == 'WAVE':
                    self._wave = [float(f) for f in value.split()]
                elif key == 'LOGW':
                    self._logw = int(value) == 1

        # since we cannot do tell() on the file after looping lines, we reopen it now
        with open(self._filename) as f:
            # search for line that indicates end of header
            while f.readline().strip() != '/':
                pass

            # store start pos
            self._spec_start = f.tell()

            # determine line length, check first 100 lines
            self._line_length = len(f.readline())
            for i in range(100):
                if len(f.readline()) != self._line_length:
                    raise ValueError('Length of lines changes.')

    def _map_lines(self):
        """Map parameters to lines in file."""

        # init params to lower limits
        params = copy.copy(self._llimits)

        # map it
        self._map_line_axis(params)

    def _map_line_axis(self, params: list, line: int = 0, axis: int = 0) -> int:
        # loop axis values
        for i in range(self._n_p[axis]):
            # get value
            val = self._llimits[axis] + i * self._steps[axis]

            # set it
            p = copy.copy(params)
            p[axis] = val

            # go deeper?
            if axis == self._n_of_dim - 1:
                # no, read lines
                self._params_line[tuple(p)] = line
                line += 1
            else:
                # go deeper
                line = self._map_line_axis(p, line, axis + 1)

        # return new line pos
        return line

    def all(self) -> List[Tuple]:
        return list(self._params_line.keys())

    def __contains__(self, params: Tuple) -> bool:
        return params in self._params_line

    def __call__(self, params: Tuple) -> Any:
        # get line in file
        line = self._params_line[params]

        # get pos in file
        pos = self._spec_start + line * self._line_length

        # open file
        with open(self._filename) as f:
            # go to position of spectrum
            f.seek(pos)

            # read line
            flux = np.array([float(v) for v in f.readline().strip().split()])

        # return spectrum
        return Spectrum(flux=flux, wave_start=self._wave[0], wave_step=self._wave[1], wave_count=self._npix,
                        wave_mode=Spectrum.Mode.LOG10LAMBDA if self._logw else Spectrum.Mode.LAMBDA)


__all__ = ['FerreGrid']



from .weight import Weight
from .fromsigma import WeightFromSigma
from .fromsnr import WeightFromSNR
from .ranges import WeightRanges
from .fromgridnearest import WeightFromGridNearest
from .fromgrid import WeightFromGrid




import logging

import numpy as np
import pandas as pd

from spexxy.data import Isochrone


def add_parser(subparsers):
    # init parser
    parser = subparsers.add_parser('apply', help='Apply isochrone to photometry')
    parser.set_defaults(func=run)

    # parameters
    parser.add_argument('isochrone', help='File containing the isochrone', type=str)
    parser.add_argument('photometry', help='File containing the photometry', type=str)
    parser.add_argument('-o', '--output', help='Output file', type=str, default='cmpiso.csv')
    parser.add_argument('--id', help='IDs from photometry file to copy to the output', nargs='+')
    parser.add_argument('--filter1', help='First filter to use', type=str, default="F606W")
    parser.add_argument('--filter2', help='Second filter to use', type=str, default="F814W")
    parser.add_argument('--filter1-iso', help='First filter in isochrone to use (--filter1 if empty)', type=str)
    parser.add_argument('--filter2-iso', help='Second filter in isochrone to use (--filter2 if empty)', type=str)
    parser.add_argument('--nearest', help='Use nearest neighbour instead of polynomial', action='store_true')
    parser.add_argument('--check-teff', help='Only apply values if Teff within range', type=float, nargs=2)
    parser.add_argument('--check-logg', help='Only apply values if logg within range', type=float, nargs=2)
    parser.add_argument('--check-mact', help='Only apply values if Mact within range', type=float, nargs=2)

    group_ex = parser.add_mutually_exclusive_group()
    group_ex.add_argument('--quadratic', help='Use quadratic polynomial instead of linear one', action='store_true')
    group_ex.add_argument('--cubic', help='Use cubic polynomial instead of linear one', action='store_true')


def run(args):
    # get isochrone
    logging.info("Loading isochrone from %s...", args.isochrone)
    isochrone = Isochrone(args.isochrone)

    # get filters in isochrone
    iso_filter1 = args.filter1.upper() if args.filter1_iso is None else args.filter1_iso.upper()
    iso_filter2 = args.filter2.upper() if args.filter2_iso is None else args.filter2_iso.upper()

    # load photometry
    logging.info("Loading photometry from %s...", args.photometry)
    photometry = pd.read_csv(args.photometry, index_col=False)

    # uppercase filters
    cols = []
    for c in photometry.columns:
        if c.endswith('_Id') or c == 'SourceNumber':
            cols.append(c)
        elif c.endswith('_err'):
            cols.append(c[:-4].upper() + '_err')
        else:
            cols.append(c.upper())
    photometry.columns = cols

    # photometry filters
    photo_filters = sorted([f for f in photometry.columns if f != 'Star_Id'])
    logging.info('Found filters in photometry: ' + ', '.join(photo_filters))
    logging.info('Using filters %s, %s in photometry.', args.filter1.upper(), args.filter2.upper())
    logging.info('Using filters %s, %s in isochrone.', iso_filter1, iso_filter2)

    # check
    if args.filter1.upper() not in photometry.columns or args.filter2.upper() not in photometry.columns:
        logging.error("Could not find given filters in photometry.")
        return

    # get interpolators
    if args.nearest:
        ip_nearest = isochrone.nearest(iso_filter1, iso_filter2)
    else:
        mode = 'quadratic' if args.quadratic else 'cubic' if args.cubic else 'linear'
        ip_teff = isochrone.interpolator('Teff', iso_filter1, iso_filter2, mode=mode)
        ip_logg = isochrone.interpolator('logg', iso_filter1, iso_filter2, mode=mode)
        ip_mact = isochrone.interpolator('M_act', iso_filter1, iso_filter2, mode=mode)
        ip_mini = isochrone.interpolator('M_ini', iso_filter1, iso_filter2, mode=mode)

    # loop all photometry
    logging.info('Applying isochrone to all stars in photometry...')
    with open(args.output, 'w') as f:
        if args.id:
            f.write(','.join(args.id) + ',')
        f.write('Teff_iso,logg_iso,Mini_iso,Mact_iso\n')

        for row in range(len(photometry)):
            # get requested ids
            ids = [photometry[i].iloc[row] for i in args.id] if args.id else []

            # get photometry
            v = photometry[args.filter1.upper()].iloc[row]
            i = photometry[args.filter2.upper()].iloc[row]
            if np.isnan(v) or np.isnan(i):
                logging.error('Magnitude is NaN for ' +
                              ', '.join([args.id[i] + '=' + str(ids[i]) for i in range(len(ids))]))
                continue

            # get values
            if args.nearest:
                tmp = ip_nearest(v - i, v)
                Teff = (tmp['teff'], None)
                logg = (tmp['logg'], None)
                Mini = (tmp['mini'], None)
                Mact = (tmp['mact'], None)
            else:
                Teff = (ip_teff(v - i, v), None)
                logg = (ip_logg(v - i, v), None)
                Mini = (ip_mini(v - i, v), None)
                Mact = (ip_mact(v - i, v), None)

            # check
            valid = True
            valid &= args.check_teff is None or args.check_teff[0] <= Teff[0] <= args.check_teff[1]
            valid &= args.check_logg is None or args.check_logg[0] <= logg[0] <= args.check_logg[1]
            valid &= args.check_mact is None or args.check_mact[0] <= Mact[0] <= args.check_mact[1]
            if not valid:
                continue

            # write results to file
            if args.id:
                f.write(','.join(['{0:d}'.format(int(i)) for i in ids]) + ',')
            f.write('{0:.4f},{1:.4f},{2:.4f},{3:.4f}\n'.format(Teff[0], logg[0], Mini[0], Mact[0]))
            f.flush()



import glob
import logging
import numpy as np
from spexxy.utils.fits import bulk_read_header

from spexxy.data import FitsSpectrum, SpectrumFits


def add_parser(subparsers):
    """
    Adds 'spexxy tellurics mean' command

    :param subparsers:  Subparser to attach new one to.
    """

    # init parser
    parser = subparsers.add_parser('mean', help='Calculates the mean tellurics from a list of fits.')
    parser.set_defaults(func=run)

    # calculate mean
    parser.add_argument('spectra', help='List of processed spectra', type=str, nargs='+')
    parser.add_argument('-o', '--output', help='File to write tellurics to', type=str, default="tellurics.fits")
    parser.add_argument('-l', '--snlimit', help='minimum SNR to use', type=float)
    parser.add_argument('-f', '--snfrac', help='if snlimit is not given, calculate it from X percent best',
                        type=float, default=10)
    parser.add_argument('-w', '--weight', help='weight by SNR', action="store_true")
    parser.add_argument('-r', '--resample', help='resample to start,step,count', nargs=3,
                        default=(None, None, None), type=float)


def run(args):
    """
    Takes a list of spectra and calculates the mean of the fitted tellurics.

    :param args:    argparse namespace with
                    .spectra:   List of files containing spectra.
                    .output:    Output file for mean tellurics
                                (default: tellurics.fits)
                    .snlimit:   Only calculate mean tellurics for spectra
                                with a S/N higher than the given number.
                    .weight:    If set, tellurics are weightes by the S/N
                                of their spectra.
    """

    # get all spectra
    filenames = []
    for s in args.spectra:
        if '*' in s:
            filenames.extend(glob.glob(s))
        else:
            filenames.append(s)

    # read headers
    logging.info('Reading all FITS headers...')
    headers = bulk_read_header(filenames, ['HIERARCH SPECTRUM SNRATIO'])

    # get all snr values
    logging.info('Extracting all S/N values...')
    snrs = {}
    for i, row in headers.iterrows():
        try:
            snrs[row['FILE']] = float(row['HIERARCH SPECTRUM SNRATIO'])
        except ValueError:
            continue

    # no snlimit given?
    if args.snlimit is not None:
        snlimit = args.snlimit
    else:
        tmp = sorted(snrs.values())
        logging.info('Calculating S/N limit from %.2f%% highest S/N values...', args.snfrac)
        snlimit = tmp[-int(len(tmp)*args.snfrac/100)]
    logging.info('Using S/N limit of %.2f...', snlimit)

    # filter by snlimit
    logging.info('Filtering spectra by S/N limit...')
    spectra = [filename for filename, snr in snrs.items() if snr > snlimit]

    # tellurics spectrum
    tellurics = None
    weights = None
    count = None
    wave_start, wave_step, wave_count = args.resample
    if wave_count is not None:
        wave_count = int(wave_count)

    # loop files
    logging.info('Processing %d spectra...', len(spectra))
    for i, spec in enumerate(spectra, 1):
        # open file
        with FitsSpectrum(spec, 'r') as fs:
            # get signal to noise
            snr = fs.header["HIERARCH SPECTRUM SNRATIO"]

            # print
            logging.info('(%d/%d) %s %-5.2f', i, len(spectra), spec, snr)

            # weight
            weight = snr if args.weight else 1.

            # some more info
            if wave_start is None:
                # WAVE extension or CRVAL/CDELT?
                if 'CRVAL1' in fs.header and 'CDELT1' in fs.header:
                    wave_start = fs.header["CRVAL1"]
                    wave_step = fs.header["CDELT1"]
                    if fs.header['CUNIT1'] == 'm':
                        wave_start *= 1e10
                        wave_step *= 1e10
                    wave_count = fs.header['NAXIS1']
                elif 'WAVE' in fs.header and fs.header['WAVE'] in fs:
                    logging.error('Combining tellurics on PIXTABLE spectra not allowed without resampling.')
                    continue
                else:
                    logging.error('Could not determine wavelength grid.')
                    continue

            # get tellurics
            tell = fs.tellurics
            if not tell:
                continue

            # resample
            tell = tell.resample(wave_start=wave_start, wave_step=wave_step, wave_count=wave_count)

            # tellurics exist? on first iteration we create the array.
            if tellurics is None:
                tellurics = np.zeros((wave_count))
                weights = np.zeros((wave_count))
                count = np.zeros((wave_count))

            # add to sum
            w = np.where(~np.isnan(tell.flux))
            tellurics[w] += weight * tell.flux[w]
            weights[w] += weight
            count[w] += 1

    # divide by sum
    w = np.where(~np.isnan(tellurics) & ~np.isnan(weights))
    tellurics[w] /= weights[w]

    # create spectrum for tellurics and save it
    tell_spec = SpectrumFits(flux=tellurics, wave_start=wave_start, wave_step=wave_step, primary=True)
    tell_spec.save(args.output)

    # output
    logging.info("Finished successfully.")



import math
from typing import Tuple
import numpy as np
import scipy.special
from numpy.polynomial import hermite

from spexxy.data.spectrum import Spectrum


class LOSVD:
    """Simple class for handling LOSVD convolution of a spectrum."""

    def __init__(self, params: Tuple):
        """Inits new LOSVD.

        Args:
            params: Parameters as (v, sig, h3, h4, h5, h6).
        """

        # init
        self._losvd = list(params)

    def kernel(self, x: np.ndarray) -> np.ndarray:
        """Creates convolution kernel for LOSVD.

        Args:
            x: x Array to create kernel on.

        Returns:
            Kernel on given x values.
        """

        # get values
        v, sig = self._losvd[:2]
        herm = self._losvd[2:]

        # helpers
        t = (x - v) / sig
        t2 = t * t

        # calculate Gaussian kernel
        k = np.exp(-0.5 * t2) / (np.sqrt(2. * np.pi) * sig)

        # Hermite polynomials normalized as in Appendix A of van der Marel &
        # Franx (1993). They are given e.g. in Appendix C of Cappellari et al. (2002)
        n = np.arange(3 + len(herm))
        nrm = np.sqrt(scipy.special.factorial(n) * 2 ** n)

        # normalize coefficients
        c = np.array([1., 0., 0.] + herm) / nrm

        # add hermites
        k *= hermite.hermval(t, c)

        # normalize kernel
        return k / np.sum(k)

    def x(self, wave_step: float) -> np.ndarray:
        """Calculates optimal X array for kernel for given spectrum.

        Args:
            wave_step: Wavelength step size in log.

        Returns:
            Numpy array containing X array.
        """

        # get values
        v, sig = self._losvd[:2]

        # wave step in km/s
        step_kms = 299792.458 * (np.exp(wave_step) - 1.)

        # get range in pixels that the kernel should cover
        pix_range = math.floor((abs(v) + 5. * sig) / step_kms)

        # create array in pixels and convert to km/s
        return np.arange(-pix_range, pix_range + 1) * step_kms

    def __call__(self, spec: Spectrum) -> np.ndarray:
        """Convolves a given spectrum with the LOSVD kernel.

        Args:
            spec: Spectrum to convolve.

        Returns:
            Convolved flux.
        """

        # spectrum in log mode?
        if spec.wave_mode != Spectrum.Mode.LOGLAMBDA:
            raise ValueError("Spectrum must be on log wavelength.")

        # get optimal X array
        x = self.x(spec.wave_step)

        # get kernel and its half-length
        k = self.kernel(x)

        # if kernel is smaller than spectrum, we can use 'same' mode, otherwise we need to cut
        if len(k) < len(spec):
            return np.convolve(spec.flux, k, mode='same')
        else:
            # convolve
            c = np.convolve(spec.flux, k, mode='full')

            # cut original range from output
            kl = len(k) // 2
            return c[kl:-kl]


__all__ = ['LOSVD']



from .interpolator import Interpolator
from .ulyss import UlyssInterpolator
from .linear import LinearInterpolator
from .tellurics import TelluricsInterpolator
from .spline import SplineInterpolator
from .sub import SubInterpolator



import logging

from ..version import version


def setup_log(name: str, filename: str = None, stream: bool = True, mode: str = 'w',
              header: bool = True) -> logging.Logger:
    """Sets up a new logger object

    Args:
        name: Name of new logger.
        filename: If given, name of file to write logs to.
        stream: If True, also log to stdout.
        mode: Open mode for log file (w/a).
        header: If True, immediately writes a spexxy header into the log.

    Returns:
        The new logger object
    """

    # get logger
    log = logging.getLogger(name)

    # remove existing handlers
    shutdown_log(name)

    # log format
    log_formatter = logging.Formatter("%(asctime)s[%(levelname)-8s]: %(message)s", datefmt='%m/%d/%Y %H:%M:%S')

    # create file handler
    if filename is not None:
        file_handler = logging.FileHandler(filename, mode=mode)
        file_handler.setFormatter(log_formatter)
        log.addHandler(file_handler)

    # stream handler as well?
    if stream:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(log_formatter)
        log.addHandler(stream_handler)

    # log header
    if header:
        log.info(r"")
        log.info(r" ___ _ __   _____  ____  ___   _ ")
        log.info(r"/ __| '_ \ / _ \ \/ /\ \/ / | | |")
        log.info(r"\__ \ |_) |  __/>  <  >  <| |_| |")
        log.info(r"|___/ .__/ \___/_/\_\/_/\_\\__, |")
        log.info(r"    | |                     __/ |")
        log.info(r"    |_|        v%3s        |___/ " % version())
        log.info(r"")
        log.info(r"        Tim-Oliver Husser        ")
        log.info(r"    thusser@uni-goettingen.de    ")
        log.info(r"")

    # return it
    return log


def shutdown_log(name: str):
    """Shuts down a logger objects, especially removes all handlers, so that we don't get a
    'too many open files' error.

    Args:
        name: Name of logger to shut down.
    """

    # get logger
    log = logging.getLogger(name)

    # remove existing handlers
    for handler in log.handlers[:]:
        log.removeHandler(handler)



import numpy as np
from typing import List, Dict

from spexxy.component import Component, TelluricsComponent
from .base import FilesRoutine


class MultiMain(FilesRoutine):
    """MultiRun iterates over a given set of main routines and runs them sequentially."""

    def __init__(self, routines: List = None, iterations: int = None, max_iterations: int = 8,
                 threshold: Dict[str, float] = None, poly_degree: int = 40, damped: bool = True,
                 factors: list = (0.7, 0.3), *args, **kwargs):
        """Initialize a new MultiMain object

        Args:
            routines: List of main routines to run.
            iterations: Number of iterations for the whole cycle.
            max_iterations: If set to a value >=2, the fit runs until it converges or until the number of iterations
                            reaches max_iterations.
            threshold: Dictionary that contains the absolute values for each fit parameter below which the fit is
                       considered as converged.
            poly_degree: Degree of Legendre polynomial used for the continuum fit.
            damped: If True the fit is repeated with a damping factor in case the fit does not converge.
            factors: List of damping factors used if the fit does not converge.
        """
        FilesRoutine.__init__(self, *args, **kwargs)

        # remember variables
        self._iterations = iterations
        self._max_iterations = max_iterations
        self._poly_degree = poly_degree
        self._threshold = threshold
        self._damped = damped
        self._factors = sorted(factors, reverse=True)

        # find main routines
        self._routines = self.get_objects(routines, FilesRoutine, 'routines')

    def parameters(self) -> List[str]:
        """Get list of parameters fitted by this routine.

        Returns:
            List of parameter names (including prefix) fitted by this routine.
        """

        # get all parameters
        parameters = []
        for routine in self._routines:
            parameters += routine.parameters()

        # make unique and sort
        return sorted(list(set(parameters)))

    def columns(self) -> List[str]:
        """Get list of columns returned by __call__.

        The returned list should include the list from parameters().

        Returns:
            List of columns returned by __call__.
        """

        # call base and add columns Iterations, Success and Convergence
        if self._max_iterations is not None and self._damped:
            return FilesRoutine.columns(self) + ['Iterations', 'Success', 'Convergence', 'Damping Factor']
        elif self._max_iterations is not None:
            return FilesRoutine.columns(self) + ['Iterations', 'Success', 'Convergence']

        return FilesRoutine.columns(self) + ['Iterations', 'Success']

    def __call__(self, filename: str) -> List[float]:
        """Process the given file.

        Args:
            filename: Name of file to process.

        Returns:
            List of final values of parameters, ordered in the same way as the return value of parameters()
        """

        # init results dict with Nones
        parameters = self.parameters()
        results = {p: None for p in parameters}

        success = []

        # list with all fit parameters
        fit_params = []
        for routine in self._routines:
            fit_params.extend(routine.fit_parameters())

        fit_params = list(set(fit_params))

        # dictionary that contains the fit results of all iteration steps, used for convergence test
        results_total = {p: [] for p in fit_params}

        # if routine checks for convergence set the threshold now
        if self._iterations is None and self._max_iterations > 1:
            # initialize thresholds
            tmp = {}
            for cmp_name, cmp in self.objects['components'].items():
                # loop over all parameters of this component
                for param_name in cmp.param_names:
                    if param_name.lower() == 'teff':
                        tmp['{} {}'.format(cmp.prefix, param_name)] = 25.
                    elif param_name.lower() == 'v' or param_name.lower() == 'sig':
                        tmp['{} {}'.format(cmp.prefix, param_name)] = 1.
                    else:
                        tmp['{} {}'.format(cmp.prefix, param_name)] = 0.05

            if self._threshold is not None:
                for cmp_name, values in self._threshold.items():
                    for param, threshold in values.items():
                        tmp['{} {}'.format(cmp_name, param)] = np.float(threshold)

            self._threshold = tmp

            maxiter = self._max_iterations
        else:
            if self._iterations is None and (self._max_iterations == 1 or self._max_iterations is None):
                maxiter = 1
            else:
                maxiter = self._iterations

            self._max_iterations = None

        # loop iterations
        for it in range(maxiter):
            self.objects['init_iter'] = {}
            for cmp_name, cmp in self.objects['components'].items():
                if isinstance(cmp, TelluricsComponent):
                    continue

                self.objects['init_iter'][cmp_name] = Component(name=cmp_name)
                for param_name in cmp.param_names:
                    self.objects['init_iter'][cmp_name].set(name=param_name, value=cmp[param_name])

            # loop main routines
            for routine in self._routines:
                # set poly degree for this routine
                routine._poly_degree = self._poly_degree

                # get parameters for this routine
                params = routine.parameters()

                # run routine
                res = routine(filename)

                # store results
                for i, p in enumerate(params):
                    # if parameter is a fit parameter in another iteration step don't overwrite previous result
                    # otherwise the error will be set to zero
                    if p in fit_params and p not in routine.fit_parameters():
                        # initialize dictionary
                        if results[p] is None:
                            results[p] = [res[i * 2], res[i * 2 + 1]]

                        continue

                    # copy both results and errors!
                    results[p] = [res[i * 2], res[i * 2 + 1]]

                # was iteration a success?
                success.append(res[-1])

            # check for convergence?
            if self._max_iterations is not None:
                # save results of all steps
                for p in fit_params:
                    results_total[p].append(results[p][0])

                # run at least for 2 iterations
                if it == 0:
                    continue

                # check for convergence
                if self.convergence(results_total):
                    # fit is successful if each iteration was a success
                    success = np.all(success)

                    # fit converged
                    converged = True

                    # convert results dict into results list
                    res = []
                    for p in parameters:
                        res.extend(results[p])

                    # add iteration, success and convergence to results
                    res.append(it + 1)
                    res.append(success)
                    res.append(converged)

                    if self._damped:
                        res.append(1.)

                    return res
                elif it == self._max_iterations - 1 and not self._damped:
                    # fit is successful if each iteration was a success
                    success = np.all(success)

                    # fit did not converge
                    converged = False

                    # convert results dict into results list
                    res = []
                    for p in parameters:
                        res.extend(results[p])

                    # add iteration, success and convergence to results
                    res.append(it + 1)
                    res.append(success)
                    res.append(converged)

                    return res

        # if fit did not converge try again with damping factor
        if self._max_iterations is not None and self._damped:
            for p in self._threshold:
                self._threshold[p] /= 3

            iterations = self._max_iterations
            for damping_factor in self._factors:
                # reset parameters to initial values
                for cmp_name, cmp in self.objects['components'].items():
                    cmp.init(filename)

                results = {p: None for p in parameters}
                success = []
                results_total = {p: [] for p in fit_params}

                # loop over fit parameters
                for it in range(3 * maxiter):
                    self.objects['init_iter'] = {}
                    for cmp_name, cmp in self.objects['components'].items():
                        if isinstance(cmp, TelluricsComponent):
                            continue

                        self.objects['init_iter'][cmp_name] = Component(name=cmp_name)
                        for param_name in cmp.param_names:
                            self.objects['init_iter'][cmp_name].set(name=param_name, value=cmp[param_name])

                    # loop main routines
                    for routine in self._routines:
                        # set poly degree for this routine
                        routine._poly_degree = self._poly_degree

                        # get parameters for this routine
                        params = routine.parameters()

                        # run routine
                        res = routine(filename)

                        # store results
                        for i, p in enumerate(params):
                            # if parameter is a fit parameter in another iteration step don't overwrite previous result
                            # otherwise the error will be set to zero
                            if p in fit_params and p not in routine.fit_parameters():
                                # initialize dictionary
                                if results[p] is None:
                                    results[p] = [res[i * 2], res[i * 2 + 1]]

                                continue

                            # copy both results and errors!
                            results[p] = [res[i * 2], res[i * 2 + 1]]

                        # calculate damped result
                        for p in params:
                            if p not in routine.fit_parameters():
                                continue

                            for cmp_name, cmp in self.objects['components'].items():
                                if not p.startswith(cmp.prefix):
                                    continue

                                name = p[len(cmp.prefix)+1:]

                                cmp[name] = (1 - damping_factor) * self.objects['init_iter'][cmp_name][
                                    name] + damping_factor * results[p][0]
                                results[p][0] = (1 - damping_factor) * self.objects['init_iter'][cmp_name][
                                    name] + damping_factor * results[p][0]

                        # was iteration a success?
                        success.append(res[-1])

                    # save results of all steps
                    for p in fit_params:
                        results_total[p].append(results[p][0])

                    # run at least for max_iterations // 2 iterations
                    if it < self._max_iterations // 2:
                        continue

                    # check for convergence
                    if self.convergence(results_total):
                        # fit is successful if each iteration was a success
                        success = np.all(success)

                        # fit converged
                        converged = True

                        # convert results dict into results list
                        res = []
                        for p in parameters:
                            res.extend(results[p])

                        # add iteration, success and convergence to results
                        res.append(iterations + it + 1)
                        res.append(success)
                        res.append(converged)
                        res.append(damping_factor)

                        return res

                iterations += 3 * self._max_iterations

            # fit is successful if each iteration was a success
            success = np.all(success)

            # fit did not converge
            converged = False

            # convert results dict into results list
            res = []
            for p in parameters:
                res.extend(results[p])

            # add iteration, success and convergence to results
            res.append(iterations)
            res.append(success)
            res.append(converged)
            res.append(self._factors[-1])

            return res

        # fit is successful if each iteration was a success
        success = np.all(success)

        # convert results dict into results list
        res = []
        for p in parameters:
            res.extend(results[p])

        # add iteration and success to results
        res.append(self._iterations)
        res.append(success)

        return res

    def convergence(self, results):
        """Returns true if the fit satisfies the convergence criteria for all fit parameters."""

        c = []
        # loop over results
        for param, res in results.items():
            # test for convergence
            c.append(abs(res[-1] - res[-2]) <= self._threshold[param])

        # return True if all parameters satisfy their convergence criterion
        return np.all(c)


__all__ = ['MultiMain']



import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def add_parser(subparsers):
    # init parser
    parser = subparsers.add_parser('cut', help='Cut isochrone')
    parser.set_defaults(func=run)

    # parameters
    parser.add_argument('input', help='Input file containing the isochrone', type=str)
    parser.add_argument('output', help='Output file', type=str)
    parser.add_argument('regions', help='Regions to include', type=str, nargs='+',
                        choices=['MS', 'SGB', 'RGB', 'HB', 'AGB'])
    parser.add_argument('-p', '--plot', help='Plot result', action='store_true')


def run(args):
    # load data
    isochrone = pd.read_csv(args.input, index_col=False, comment='#')

    # sort isochrone by M_ini
    isochrone.sort_values('M_ini', inplace=True)

    # plot isochrone?
    if args.plot:
        plt.plot(isochrone['Teff'], isochrone['logg'], label='Input isochrone')

    # find MS by finding first turn-around for Teff
    Teff = isochrone['Teff']
    MS = None
    for i in range(len(Teff) - 1):
        if Teff.iloc[i+1] < Teff.iloc[i]:
            logging.info('Found main sequence turnoff at %.2fK.', Teff.iloc[i])
            MS = isochrone.iloc[:i+1]
            isochrone = isochrone.iloc[i+1:]
            break

    # find tip of RGB by finding first turn-off for Teff at logg<1
    GB = None
    logg = isochrone['logg']
    Teff = isochrone['Teff']
    for i in range(len(Teff) - 1):
        if logg.iloc[i] < 1 and Teff.iloc[i + 1] > Teff.iloc[i]:
            logging.info('Found tip of RPG at %.2fK.', Teff.iloc[i])
            GB = isochrone.iloc[:i + 1]
            isochrone = isochrone.iloc[i + 1:]
            break

    # find HB by finding highest temperature
    Teff = isochrone['Teff']
    i = np.argmax(Teff.values)
    logging.info('Found end of horizontal branch at %.2fK.', Teff.iloc[i])
    HB = isochrone.iloc[i:]
    isochrone = isochrone.iloc[:i]

    # plot parts
    if args.plot:
        # plots
        plt.scatter(MS['Teff'], MS['logg'], label='MS', c='blue')
        plt.scatter(GB['Teff'], GB['logg'], label='SGB/RGB', c='red')
        plt.scatter(HB['Teff'], HB['logg'], label='HB', c='black')
        #plt.scatter(AGB['Teff'], AGB['logg'], label='AGB')
        # stuff
        plt.gca().invert_xaxis()
        plt.gca().invert_yaxis()
        plt.grid()
        plt.legend(loc=7)
        plt.xlabel('Teff [K]')
        plt.ylabel('logg [cm/s/s]')
        plt.show()

    # connect parts again
    parts = []
    if 'MS' in args.regions:
        parts += [MS]
    if 'SGB' in args.regions or 'RGB' in args.regions:
        parts += [GB]
    if 'HB' in args.regions:
        parts += [HB]
    #if 'AGB' in args.regions:
    #    parts += [AGB]
    output = pd.concat(parts)

    # write output
    output.to_csv(args.output, index=False)



from .grid import Grid, GridAxis
from .files import FilesGrid
from .values import ValuesGrid
from .synspec import SynspecGrid
from .ferre import FerreGrid
from .fits import FitsGrid
from .sub import SubGrid



import copy

import numpy as np
from typing import List, Tuple

from . import Interpolator
from ..grid import Grid, GridAxis
from ..data import Spectrum


def calc_2nd_derivs_spline(x: list, y: list, yp1=np.inf, ypn=np.inf):
    """Calculates the 2nd derivatives for a spline.
    Python conversion from the C++ code in chapter "Cubic Spline Interpolation" in the
    "Numerical Recipes in C++, 2nd Edition".

    Args:
        x: Input x values.
        y: Input y values.
        yp1: First derivative at point 0. If set to np.inf, use natural boundary condition and set 2nd deriv to 0.
        ypn: First derivative at point n-1. np.inf means the same as for yp1.

    Returns:
        Second derivates for all points given by x and y.
    """

    # get number of elements
    n = len(x)

    # create arrays for u and 2nd derivs
    if hasattr(y[0], '__iter__'):
        y2 = np.zeros((n, len(y[0])))
        u = np.zeros((n, len(y[0])))
    else:
        y2 = np.zeros((n))
        u = np.zeros((n))

    # derivatives for point 0 given?
    if not np.isinf(yp1):
        y2[0] += -0.5
        u[0] += (3. / (x[1] - x[0])) * ((y[1] - y[0]) / (x[1] - x[0]) - yp1)

    # decomposition loop of the tridiagonal algorithm
    for i in range(1, n - 1):
        sig = (x[i] - x[i - 1]) / (x[i + 1] - x[i - 1])
        p = sig * y2[i - 1] + 2.0
        y2[i] = (sig - 1.0) / p
        u[i] = (y[i + 1] - y[i]) / (x[i + 1] - x[i]) - (y[i] - y[i - 1]) / (x[i] - x[i - 1])
        u[i] = (6.0 * u[i] / (x[i + 1] - x[i - 1]) - sig * u[i - 1]) / p

    # derivatives for point n-1 given?
    if not np.isinf(ypn):
        qn = 0.5
        un = (3. / (x[n - 1] - x[n - 2])) * (ypn - (y[n - 1] - y[n - 2]) / (x[n - 1] - x[n - 2]))
        y2[n - 1] += (un - qn * u[n - 2]) / (qn * y2[n - 2] + 1.)

    # backsubstitution loop of the tridiagonal algorithm
    for k in range(n - 2, 0, -1):
        y2[k] = y2[k] * y2[k + 1] + u[k]

    # finished
    return y2


class SplineInterpolator(Interpolator):
    """A cubic spline interpolator that operates on a given grid."""
    def __init__(self, grid: Grid, derivs: Grid = None, n: int = 1, verbose: bool = False, *args, **kwargs):
        """Initializes a new linear interpolator.

        Args:
            grid: Grid to interpolate on.
            derivs: If given, contains a second grid at the same parameters as grid, but containg 2nd derivatives for
                the first axis of the grid.
            n: Number of points on each side to use for calculating derivatives.
            verbose: If True, output some more logs
        """
        Interpolator.__init__(self, *args, **kwargs)

        # grids
        self.log.info('Initializing spline interpolator...')
        self._grid: Grid = self.get_objects(grid, Grid, 'grids', single=True)
        self._derivs: Grid = self.get_objects(derivs, Grid, 'grids', single=True)

        # init
        self._axes = self._grid.axes()
        self._npoints = n
        self._verbose = verbose

    @property
    def grid(self) -> Grid:
        """Returns grid used in this interpolator.

        Returns:
            Grid used for this interpolator
        """
        return self._grid

    def axes(self) -> List[GridAxis]:
        """Returns information about the axes.

        Returns:
            List of GridAxis objects describing the grid's axes.
        """
        return self._axes

    def __call__(self, params: Tuple) -> Spectrum:
        """Interpolates at the given parameter set.

        Args:
            params: Parameter set to interpolate at.

        Returns:
            Interpolated spectrum at given position.
        """
        return self._interpolate(tuple(params))

    def _interpolate(self, params: Tuple, axis: int = None):
        """Do the actual interpolation at the given axis, then continue for axis-1.

        Args:
            params: Parameter set to interpolate at.
            axis: Axis to interpolate in - if None, we start at last available axis.

        Returns:
            Interpolated spectrum at given position.
        """

        # no axis given, start at latest
        if axis is None:
            axis = len(self._axes) - 1

        # caching?
        if self.cache_level >= len(self._axes) - axis and params in self.cache:
            return copy.deepcopy(self.cache[params])

        # check boundaries
        if params[axis] < self._axes[axis].min or params[axis] > self._axes[axis].max:
            raise KeyError('Requested parameters are outside the grid.')

        # let's get all possible values for the given axis
        axisValues = self._grid.axis_values(axis)

        # if params[axis] is on axis; return it directly
        if params[axis] in axisValues:
            if axis == 0 and tuple(params) in self._grid:
                # if this is the first axis AND params are in grid, return it
                return self._grid(tuple(params))
            elif axis > 0:
                # if it's another axis, interpolate
                return self._interpolate(tuple(params), axis - 1)
            # if it's neither, just continue

        # find the next lower and the next higher axis value
        # on the last axis we enforce that the neighbour must exist
        p_lower = self._grid.neighbour(tuple(params), axis, 0, must_exist=axis == 0)
        p_higher = self._grid.neighbour(tuple(params), axis, 1, must_exist=axis == 0)
        if p_lower is None or p_higher is None:
            raise KeyError('No direct neighbours found in grid.')

        # get axis values
        x_lower = p_lower[axis]
        x_higher = p_higher[axis]

        # get data for p_lower and p_higher
        if axis == 0:
            lower_data = self._grid(p_lower)
            higher_data = self._grid(p_higher)
        else:
            lower_data = self._interpolate(p_lower)
            higher_data = self._interpolate(p_higher)

        # do we have pre-calculated 2nd derivatives?
        if axis == 0 and self._derivs is not None:
            # set them
            lower_deriv = self._derivs(p_lower)
            higher_deriv = self._derivs(p_higher)

        else:
            # get data for a maximum of N lower and higher axis values
            x = []
            y = []
            ilower = None
            for i in range(-self._npoints, 2+self._npoints):
                if i == 0:
                    # we got this value already, store it
                    x.append(x_lower)
                    y.append(lower_data)
                    # set ilower index to this index
                    ilower = len(x) - 1
                elif i == 1:
                    # we got this value already, store it
                    x.append(x_higher)
                    y.append(higher_data)
                else:
                    # try to fetch additional data
                    try:
                        # find neighbour
                        p = self._grid.neighbour(tuple(params), axis, i, must_exist=axis == 0)
                        # found one?
                        if p is not None:
                            # in grid or do we need to interpolate?
                            if p in self._grid:
                                y.append(self._grid(p))
                            else:
                                y.append(self._interpolate(p))
                            x.append(p[axis])
                    except KeyError:
                        pass

            # calculate 2nd derivatives
            if self._verbose:
                self.log.info('Creating 2nd derivates for axis %s with values %s.', self._grid.axis_name(axis), x)
            y2 = calc_2nd_derivs_spline(x, y)

            # set them
            lower_deriv = y2[ilower]
            higher_deriv = y2[ilower + 1]

        # log
        if self._verbose:
            self.log.info('Interpolate for axis %s at %.2f from neighbours at %.2f and %.2f.',
                          self._grid.axis_name(axis), params[axis], x_lower, x_higher)
            self.log.info('2nd derivation at 1st point is %.2g for next lower and %.2g for next higher neighbour.',
                          lower_deriv[0], higher_deriv[1])

        # calculate coefficients
        h = (x_higher - x_lower)
        A = (x_higher - params[axis]) / h
        B = (params[axis] - x_lower) / h
        C = (A**3 - A) * h**2 / 6.
        D = (B**3 - B) * h**2 / 6.
        if self._verbose:
            self.log.info('A=%.2f, B=%.2f, C=%.2f, D=%.2f', A, B, C, D)

        # interpolate
        ip = lower_data * A + higher_data * B + lower_deriv * C + higher_deriv * D

        # caching?
        if self.cache_level >= len(self._axes) - axis:
            self.cache[params] = copy.deepcopy(ip)

        # finished
        return ip


__all__ = ['SplineInterpolator', 'calc_2nd_derivs_spline']



import logging
import argparse
from typing import List

import yaml

from spexxy.data import Spectrum, SpectrumFits
from spexxy.grid import Grid
from spexxy.interpolator import Interpolator
from spexxy.object import create_object


log = logging.getLogger(__name__)


def add_parser(subparsers):
    # init parser
    parser = subparsers.add_parser('info', help='Prints information about a given grid',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('config', type=str, help='Config for grid')
    parser.add_argument('-v', '--values', action='store_true', help='Show all values on axes')

    # argparse wrapper for create_grid
    parser.set_defaults(func=lambda args: get_spectrum_from_grid(**vars(args)))


def get_spectrum_from_grid(config: str, values: bool, **kwargs):
    """Prints info about grid given in config.

    Args:
        config: Config to build grid from.
        values: If True, print all values per axis.
    """

    # create grid
    with open(config, 'r') as f:
        cfg = yaml.safe_load(f)
    grid = create_object(config=cfg, log=log)
    if not isinstance(grid, Grid) and not isinstance(grid, Interpolator):
        raise ValueError('Object from config is neither a Grid nor an Interpolator.')

    # get axes
    axes = grid.axes()
    print('Found %d axes: %s' % (len(axes), ' '.join([ax.name for ax in axes])))

    # loop axes
    for i, ax in enumerate(axes, 1):
        # print axis name
        print('%3d. %s' % (i, ax.name))

        # values?
        if ax.values is not None:
            print('     # steps: %d' % len(ax.values))
            if values:
                print('     values:  %s' % ','.join(['%.2f' % v for v in ax.values]))
        else:
            print('     continuous')

        # min/max?
        if ax.min is not None and ax.max is not None:
            print('     min/max: %.2f / %.2f' % (ax.min, ax.max))

        # initial
        if ax.initial is not None:
            print('     initial: %.2f' % ax.initial)

